# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

# pylint: disable=consider-using-enumerate

import copy
import ctypes

from collections.abc import Iterable

import awkward as ak
from awkward._v2.index import Index
from awkward._v2.index import Index8
from awkward._v2.index import Index64
from awkward._v2._slicing import NestedIndexError
from awkward._v2.contents.content import Content
from awkward._v2.forms.unionform import UnionForm
from awkward._v2.forms.form import _parameters_equal

np = ak.nplike.NumpyMetadata.instance()
numpy = ak.nplike.Numpy.instance()


class UnionArray(Content):
    is_UnionType = True

    def __init__(
        self, tags, index, contents, identifier=None, parameters=None, nplike=None
    ):
        if not (isinstance(tags, Index) and tags.dtype == np.dtype(np.int8)):
            raise TypeError(
                "{} 'tags' must be an Index with dtype=int8, not {}".format(
                    type(self).__name__, repr(tags)
                )
            )
        if not isinstance(index, Index) and index.dtype in (
            np.dtype(np.int32),
            np.dtype(np.uint32),
            np.dtype(np.int64),
        ):
            raise TypeError(
                "{} 'index' must be an Index with dtype in (int32, uint32, int64), "
                "not {}".format(type(self).__name__, repr(index))
            )

        if not isinstance(contents, Iterable):
            raise TypeError(
                "{} 'contents' must be iterable, not {}".format(
                    type(self).__name__, repr(contents)
                )
            )
        if not isinstance(contents, list):
            contents = list(contents)

        for content in contents:
            if not isinstance(content, Content):
                raise TypeError(
                    "{} all 'contents' must be Content subclasses, not {}".format(
                        type(self).__name__, repr(content)
                    )
                )

        if (
            tags.nplike.known_shape
            and index.nplike.known_shape
            and tags.length > index.length
        ):
            raise ValueError(
                "{} len(tags) ({}) must be <= len(index) ({})".format(
                    type(self).__name__, tags.length, index.length
                )
            )
        if nplike is None:
            for content in contents:
                if nplike is None:
                    nplike = content.nplike
                    break
                elif nplike is not content.nplike:
                    raise TypeError(
                        "{} 'contents' must use the same array library (nplike): {} vs {}".format(
                            type(self).__name__,
                            type(nplike).__name__,
                            type(content.nplike).__name__,
                        )
                    )
        if nplike is None:
            nplike = tags.nplike

        self._tags = tags
        self._index = index
        self._contents = contents
        self._init(identifier, parameters, nplike)

    @property
    def tags(self):
        return self._tags

    @property
    def index(self):
        return self._index

    def content(self, index):
        return self._contents[index]

    @property
    def contents(self):
        return self._contents

    Form = UnionForm

    def _form_with_key(self, getkey):
        form_key = getkey(self)
        return self.Form(
            self._tags.form,
            self._index.form,
            [x._form_with_key(getkey) for x in self._contents],
            has_identifier=self._identifier is not None,
            parameters=self._parameters,
            form_key=form_key,
        )

    def _to_buffers(self, form, getkey, container, nplike):
        assert isinstance(form, self.Form)
        key1 = getkey(self, form, "tags")
        key2 = getkey(self, form, "index")
        container[key1] = ak._v2._util.little_endian(self._tags.raw(nplike))
        container[key2] = ak._v2._util.little_endian(self._index.raw(nplike))
        for i, content in enumerate(self._contents):
            content._to_buffers(form.content(i), getkey, container, nplike)

    @property
    def typetracer(self):
        tt = ak._v2._typetracer.TypeTracer.instance()
        return UnionArray(
            ak._v2.index.Index(self._tags.raw(tt)),
            ak._v2.index.Index(self._index.raw(tt)),
            [x.typetracer for x in self._contents],
            self._typetracer_identifier(),
            self._parameters,
            tt,
        )

    @property
    def length(self):
        return self._tags.length

    def _forget_length(self):
        return UnionArray(
            self._tags.forget_length(),
            self._index,
            self._contents,
            self._identifier,
            self._parameters,
            self._nplike,
        )

    def __repr__(self):
        return self._repr("", "", "")

    def _repr(self, indent, pre, post):
        out = [indent, pre, "<UnionArray len="]
        out.append(repr(str(self.length)))
        out.append(">")
        out.extend(self._repr_extra(indent + "    "))
        out.append("\n")
        out.append(self._tags._repr(indent + "    ", "<tags>", "</tags>\n"))
        out.append(self._index._repr(indent + "    ", "<index>", "</index>\n"))

        for i, x in enumerate(self._contents):
            out.append(f"{indent}    <content index={repr(str(i))}>\n")
            out.append(x._repr(indent + "        ", "", "\n"))
            out.append(f"{indent}    </content>\n")

        out.append(indent + "</UnionArray>")
        out.append(post)
        return "".join(out)

    def merge_parameters(self, parameters):
        return UnionArray(
            self._tags,
            self._index,
            self._contents,
            self._identifier,
            ak._v2._util.merge_parameters(self._parameters, parameters),
            self._nplike,
        )

    def _getitem_nothing(self):
        return self._getitem_range(slice(0, 0))

    def _getitem_at(self, where):
        if not self._nplike.known_data:
            return ak._v2._typetracer.OneOf(
                [x._getitem_at(where) for x in self._contents]
            )

        if where < 0:
            where += self.length
        if self._nplike.known_shape and not 0 <= where < self.length:
            raise NestedIndexError(self, where)
        tag, index = self._tags[where], self._index[where]
        return self._contents[tag]._getitem_at(index)

    def _getitem_range(self, where):
        if not self._nplike.known_shape:
            return self

        start, stop, step = where.indices(self.length)
        assert step == 1
        return UnionArray(
            self._tags[start:stop],
            self._index[start:stop],
            self._contents,
            self._range_identifier(start, stop),
            self._parameters,
            self._nplike,
        )

    def _getitem_field(self, where, only_fields=()):
        return UnionArray(
            self._tags,
            self._index,
            [x._getitem_field(where, only_fields) for x in self._contents],
            self._field_identifier(where),
            None,
            self._nplike,
        )

    def _getitem_fields(self, where, only_fields=()):
        return UnionArray(
            self._tags,
            self._index,
            [x._getitem_fields(where, only_fields) for x in self._contents],
            self._fields_identifer(where),
            None,
            self._nplike,
        )

    def _carry(self, carry, allow_lazy, exception):
        assert isinstance(carry, ak._v2.index.Index)

        try:
            nexttags = self._tags[carry.data]
            nextindex = self._index[: self._tags.length][carry.data]
        except IndexError as err:
            if issubclass(exception, NestedIndexError):
                raise exception(self, carry.data, str(err))
            else:
                raise exception(str(err))

        return UnionArray(
            nexttags,
            nextindex,
            self._contents,
            self._carry_identifier(carry, exception),
            self._parameters,
            self._nplike,
        )

    def project(self, index):
        lentags = self._tags.length
        assert not self._index.length < lentags
        lenout = ak._v2.index.Index64.empty(1, self._nplike)
        tmpcarry = ak._v2.index.Index64.empty(lentags, self._nplike)
        assert (
            lenout.nplike is self._nplike
            and tmpcarry.nplike is self._nplike
            and self._tags.nplike is self._nplike
            and self._index.nplike is self._nplike
        )
        self._handle_error(
            self._nplike[
                "awkward_UnionArray_project",
                lenout.dtype.type,
                tmpcarry.dtype.type,
                self._tags.dtype.type,
                self._index.dtype.type,
            ](
                lenout.data,
                tmpcarry.data,
                self._tags.data,
                self._index.data,
                lentags,
                index,
            )
        )
        nextcarry = ak._v2.index.Index64(tmpcarry.data[: lenout[0]], self._nplike)
        return self._contents[index]._carry(nextcarry, False, NestedIndexError)

    @staticmethod
    def regular_index(tags, IndexClass=Index64, nplike=None):
        if nplike is None:
            nplike = ak.nplike.of(tags)

        lentags = tags.length
        size = ak._v2.index.Index64.empty(1, nplike)
        assert size.nplike is nplike and tags.nplike is nplike
        Content._selfless_handle_error(
            nplike[
                "awkward_UnionArray_regular_index_getsize",
                size.dtype.type,
                tags.dtype.type,
            ](
                size.data,
                tags.data,
                lentags,
            )
        )
        current = IndexClass.empty(size[0], nplike)
        outindex = IndexClass.empty(lentags, nplike)
        assert (
            outindex.nplike is nplike
            and current.nplike is nplike
            and tags.nplike is nplike
        )
        Content._selfless_handle_error(
            nplike[
                "awkward_UnionArray_regular_index",
                outindex.dtype.type,
                current.dtype.type,
                tags.dtype.type,
            ](
                outindex.data,
                current.data,
                size[0],
                tags.data,
                lentags,
            )
        )
        return outindex

    def _regular_index(self, tags):
        return self.regular_index(
            tags, IndexClass=type(self._index), nplike=self._nplike
        )

    @staticmethod
    def nested_tags_index(
        offsets, counts, TagsClass=Index8, IndexClass=Index64, nplike=None
    ):
        if nplike is None:
            nplike = ak.nplike.of(offsets, counts)

        f_offsets = ak._v2.index.Index64(copy.deepcopy(offsets.data))
        contentlen = f_offsets[f_offsets.length - 1]

        tags = TagsClass.empty(contentlen, nplike)
        index = IndexClass.empty(contentlen, nplike)

        for tag, count in enumerate(counts):
            assert (
                tags.nplike is nplike
                and index.nplike is nplike
                and f_offsets.nplike is nplike
                and count.nplike is nplike
            )
            Content._selfless_handle_error(
                nplike[
                    "awkward_UnionArray_nestedfill_tags_index",
                    tags.dtype.type,
                    index.dtype.type,
                    f_offsets.dtype.type,
                    count.dtype.type,
                ](
                    tags.data,
                    index.data,
                    f_offsets.data,
                    tag,
                    count.data,
                    f_offsets.length - 1,
                )
            )
        return (tags, index)

    def _nested_tags_index(self, offsets, counts):
        return self.nested_tags_index(
            offsets,
            counts,
            TagsClass=type(self._tags),
            IndexClass=type(self._index),
            nplike=self._nplike,
        )

    def _getitem_next_jagged_generic(self, slicestarts, slicestops, slicecontent, tail):
        simplified = self.simplify_uniontype()
        if isinstance(simplified, ak._v2.contents.UnionArray):
            raise NestedIndexError(
                self,
                ak._v2.contents.ListArray(
                    slicestarts, slicestops, slicecontent, None, None, self._nplike
                ),
                "cannot apply jagged slices to irreducible union arrays",
            )
        return simplified._getitem_next_jagged(
            slicestarts, slicestops, slicecontent, tail
        )

    def _getitem_next_jagged(self, slicestarts, slicestops, slicecontent, tail):
        return self._getitem_next_jagged_generic(
            slicestarts, slicestops, slicecontent, tail
        )

    def _getitem_next(self, head, tail, advanced):
        if head == ():
            return self

        elif isinstance(
            head, (int, slice, ak._v2.index.Index64, ak._v2.contents.ListOffsetArray)
        ):
            outcontents = []
            for i in range(len(self._contents)):
                projection = self.project(i)
                outcontents.append(projection._getitem_next(head, tail, advanced))
            outindex = self._regular_index(self._tags)

            out = UnionArray(
                self._tags,
                outindex,
                outcontents,
                self._identifier,
                self._parameters,
                self._nplike,
            )
            return out.simplify_uniontype()

        elif ak._util.isstr(head):
            return self._getitem_next_field(head, tail, advanced)

        elif isinstance(head, list):
            return self._getitem_next_fields(head, tail, advanced)

        elif head is np.newaxis:
            return self._getitem_next_newaxis(tail, advanced)

        elif head is Ellipsis:
            return self._getitem_next_ellipsis(tail, advanced)

        elif isinstance(head, ak._v2.contents.IndexedOptionArray):
            return self._getitem_next_missing(head, tail, advanced)

        else:
            raise AssertionError(repr(head))

    def simplify_uniontype(self, merge=True, mergebool=False):
        if self._nplike.known_shape and self._index.length < self._tags.length:
            raise ValueError("invalid UnionArray: len(index) < len(tags)")

        length = self._tags.length
        tags = ak._v2.index.Index8.empty(length, self._nplike)
        index = ak._v2.index.Index64.empty(length, self._nplike)
        contents = []

        for i, self_cont in enumerate(self._contents):
            if isinstance(self_cont, UnionArray):
                innertags = self_cont._tags
                innerindex = self_cont._index
                innercontents = self_cont._contents

                for j, inner_cont in enumerate(innercontents):
                    unmerged = True
                    for k in range(len(contents)):
                        if merge and contents[k].mergeable(inner_cont, mergebool):
                            assert (
                                tags.nplike is self._nplike
                                and index.nplike is self._nplike
                                and self._tags.nplike is self._nplike
                                and self._index.nplike is self._nplike
                                and innertags.nplike is self._nplike
                                and innerindex.nplike is self._nplike
                            )
                            self._handle_error(
                                self._nplike[
                                    "awkward_UnionArray_simplify",
                                    tags.dtype.type,
                                    index.dtype.type,
                                    self._tags.dtype.type,
                                    self._index.dtype.type,
                                    innertags.dtype.type,
                                    innerindex.dtype.type,
                                ](
                                    tags.data,
                                    index.data,
                                    self._tags.data,
                                    self._index.data,
                                    innertags.data,
                                    innerindex.data,
                                    k,
                                    j,
                                    i,
                                    length,
                                    contents[k].length,
                                )
                            )
                            contents[k] = contents[k].merge(inner_cont)
                            unmerged = False
                            break

                    if unmerged:
                        assert (
                            tags.nplike is self._nplike
                            and index.nplike is self._nplike
                            and self._tags.nplike is self._nplike
                            and self._index.nplike is self._nplike
                            and innertags.nplike is self._nplike
                            and innerindex.nplike is self._nplike
                        )
                        self._handle_error(
                            self._nplike[
                                "awkward_UnionArray_simplify",
                                tags.dtype.type,
                                index.dtype.type,
                                self._tags.dtype.type,
                                self._index.dtype.type,
                                innertags.dtype.type,
                                innerindex.dtype.type,
                            ](
                                tags.data,
                                index.data,
                                self._tags.data,
                                self._index.data,
                                innertags.data,
                                innerindex.data,
                                len(contents),
                                j,
                                i,
                                length,
                                0,
                            )
                        )
                        contents.append(inner_cont)

            else:
                unmerged = True
                for k in range(len(contents)):
                    if contents[k] is self_cont:
                        assert (
                            tags.nplike is self._nplike
                            and index.nplike is self._nplike
                            and self._tags.nplike is self._nplike
                            and self._index.nplike is self._nplike
                        )
                        self._handle_error(
                            self._nplike[
                                "awkward_UnionArray_simplify_one",
                                tags.dtype.type,
                                index.dtype.type,
                                self._tags.dtype.type,
                                self._index.dtype.type,
                            ](
                                tags.data,
                                index.data,
                                self._tags.data,
                                self._index.data,
                                k,
                                i,
                                length,
                                0,
                            )
                        )
                        unmerged = False
                        break

                    elif merge and contents[k].mergeable(self_cont, mergebool):
                        assert (
                            tags.nplike is self._nplike
                            and index.nplike is self._nplike
                            and self._tags.nplike is self._nplike
                            and self._index.nplike is self._nplike
                        )
                        self._handle_error(
                            self._nplike[
                                "awkward_UnionArray_simplify_one",
                                tags.dtype.type,
                                index.dtype.type,
                                self._tags.dtype.type,
                                self._index.dtype.type,
                            ](
                                tags.data,
                                index.data,
                                self._tags.data,
                                self._index.data,
                                k,
                                i,
                                length,
                                contents[k].length,
                            )
                        )
                        contents[k] = contents[k].merge(self_cont)
                        unmerged = False
                        break

                if unmerged:
                    assert (
                        tags.nplike is self._nplike
                        and index.nplike is self._nplike
                        and self._tags.nplike is self._nplike
                        and self._index.nplike is self._nplike
                    )
                    self._handle_error(
                        self._nplike[
                            "awkward_UnionArray_simplify_one",
                            tags.dtype.type,
                            index.dtype.type,
                            self._tags.dtype.type,
                            self._index.dtype.type,
                        ](
                            tags.data,
                            index.data,
                            self._tags.data,
                            self._index.data,
                            len(contents),
                            i,
                            length,
                            0,
                        )
                    )
                    contents.append(self_cont)

        if len(contents) > 2**7:
            raise NotImplementedError(
                "FIXME: handle UnionArray with more than 127 contents"
            )

        if len(contents) == 1:
            return contents[0]._carry(index, True, NestedIndexError)

        else:
            return UnionArray(
                tags, index, contents, self._identifier, self._parameters, self._nplike
            )

    def num(self, axis, depth=0):
        posaxis = self.axis_wrap_if_negative(axis)
        if posaxis == depth:
            out = ak._v2.index.Index64.empty(1, self.nplike)
            out[0] = self.length
            return ak._v2.contents.numpyarray.NumpyArray(out, None, None, self._nplike)[
                0
            ]
        else:
            contents = []
            for content in self._contents:
                contents.append(content.num(posaxis, depth))
            out = UnionArray(
                self._tags, self._index, contents, None, self._parameters, self._nplike
            )
            return out.simplify_uniontype(True, False)

    def _offsets_and_flattened(self, axis, depth):
        posaxis = self.axis_wrap_if_negative(axis)

        if posaxis == depth:
            raise np.AxisError(self, "axis=0 not allowed for flatten")

        else:
            has_offsets = False
            offsetsraws = self._nplike.empty(len(self._contents), dtype=np.intp)
            contents = []

            for i in range(len(self._contents)):
                offsets, flattened = self._contents[i]._offsets_and_flattened(
                    posaxis, depth
                )
                offsetsraws[i] = offsets.ptr
                contents.append(flattened)
                has_offsets = offsets.length != 0

            if has_offsets:
                total_length = ak._v2.index.Index64.empty(1, self._nplike)
                assert (
                    total_length.nplike is self._nplike
                    and self._tags.nplike is self._nplike
                    and self._index.nplike is self._nplike
                )
                self._handle_error(
                    self._nplike[
                        "awkward_UnionArray_flatten_length",
                        total_length.dtype.type,
                        self._tags.dtype.type,
                        self._index.dtype.type,
                        np.int64,
                    ](
                        total_length.data,
                        self._tags.data,
                        self._index.data,
                        self._tags.length,
                        offsetsraws.ctypes.data_as(
                            ctypes.POINTER(ctypes.POINTER(ctypes.c_int64))
                        ),
                    )
                )

                totags = ak._v2.index.Index8.empty(total_length[0], self._nplike)
                toindex = ak._v2.index.Index64.empty(total_length[0], self._nplike)
                tooffsets = ak._v2.index.Index64.empty(
                    self._tags.length + 1, self._nplike
                )

                assert (
                    totags.nplike is self._nplike
                    and toindex.nplike is self._nplike
                    and tooffsets.nplike is self._nplike
                    and self._tags.nplike is self._nplike
                    and self._index.nplike is self._nplike
                )
                self._handle_error(
                    self._nplike[
                        "awkward_UnionArray_flatten_combine",
                        totags.dtype.type,
                        toindex.dtype.type,
                        tooffsets.dtype.type,
                        self._tags.dtype.type,
                        self._index.dtype.type,
                        np.int64,
                    ](
                        totags.data,
                        toindex.data,
                        tooffsets.data,
                        self._tags.data,
                        self._index.data,
                        self._tags.length,
                        offsetsraws.ctypes.data_as(
                            ctypes.POINTER(ctypes.POINTER(ctypes.c_int64))
                        ),
                    )
                )

                return (
                    tooffsets,
                    UnionArray(
                        totags, toindex, contents, None, self._parameters, self._nplike
                    ),
                )

            else:
                offsets = ak._v2.index.Index64.zeros(1, self._nplike, dtype=np.int64)
                return (
                    offsets,
                    UnionArray(
                        self._tags,
                        self._index,
                        contents,
                        None,
                        self._parameters,
                        self._nplike,
                    ),
                )

    def mergeable(self, other, mergebool):
        if not _parameters_equal(self._parameters, other._parameters):
            return False
        return True

    def merging_strategy(self, others):
        if len(others) == 0:
            raise ValueError(
                "to merge this array with 'others', at least one other must be provided"
            )

        head = [self]
        tail = []

        for i in range(len(others)):
            head.append(others[i])

        if any(
            isinstance(x.nplike, ak._v2._typetracer.TypeTracer) for x in head + tail
        ):
            head = [
                x
                if isinstance(x.nplike, ak._v2._typetracer.TypeTracer)
                else x.typetracer
                for x in head
            ]
            tail = [
                x
                if isinstance(x.nplike, ak._v2._typetracer.TypeTracer)
                else x.typetracer
                for x in tail
            ]

        return (head, tail)

    def _reverse_merge(self, other):
        theirlength = other.length
        mylength = self.length

        tags = ak._v2.index.Index8.empty(theirlength + mylength, self._nplike)
        index = ak._v2.index.Index64.empty(theirlength + mylength, self._nplike)

        contents = [other]
        contents.extend(self.contents)

        assert tags.nplike is self._nplike
        self._handle_error(
            self._nplike["awkward_UnionArray_filltags_const", tags.dtype.type](
                tags.data,
                0,
                theirlength,
                0,
            )
        )

        assert index.nplike is self._nplike
        self._handle_error(
            self._nplike["awkward_UnionArray_fillindex_count", index.dtype.type](
                index.data,
                0,
                theirlength,
            )
        )

        assert tags.nplike is self._nplike and self.tags.nplike is self._nplike
        self._handle_error(
            self._nplike[
                "awkward_UnionArray_filltags",
                tags.dtype.type,
                self.tags.dtype.type,
            ](
                tags.data,
                theirlength,
                self.tags.data,
                mylength,
                1,
            )
        )

        assert index.nplike is self._nplike and self.index.nplike is self._nplike
        self._handle_error(
            self._nplike[
                "awkward_UnionArray_fillindex",
                index.dtype.type,
                self.index.dtype.type,
            ](
                index.data,
                theirlength,
                self.index.data,
                mylength,
            )
        )

        if len(contents) > 2**7:
            raise AssertionError("FIXME: handle UnionArray with more than 127 contents")

        parameters = ak._v2._util.merge_parameters(self._parameters, other._parameters)
        return ak._v2.contents.unionarray.UnionArray(
            tags, index, contents, None, parameters, self._nplike
        )

    def mergemany(self, others):
        if len(others) == 0:
            return self

        head, tail = self._merging_strategy(others)

        total_length = 0
        for array in head:
            total_length += array.length

        nexttags = ak._v2.index.Index8.empty(total_length, self._nplike)
        nextindex = ak._v2.index.Index64.empty(total_length, self._nplike)

        nextcontents = []
        length_so_far = 0
        parameters = self._parameters

        for array in head:
            parameters = ak._v2._util.merge_parameters(
                self._parameters, array._parameters, True
            )
            if isinstance(array, ak._v2.contents.unionarray.UnionArray):
                union_tags = ak._v2.index.Index(array.tags)
                union_index = ak._v2.index.Index(array.index)
                union_contents = array.contents
                assert (
                    nexttags.nplike is self._nplike
                    and union_tags.nplike is self._nplike
                )
                self._handle_error(
                    self._nplike[
                        "awkward_UnionArray_filltags",
                        nexttags.dtype.type,
                        union_tags.dtype.type,
                    ](
                        nexttags.data,
                        length_so_far,
                        union_tags.data,
                        array.length,
                        len(nextcontents),
                    )
                )
                assert (
                    nextindex.nplike is self._nplike
                    and union_index.nplike is self._nplike
                )
                self._handle_error(
                    self._nplike[
                        "awkward_UnionArray_fillindex",
                        nextindex.dtype.type,
                        union_index.dtype.type,
                    ](
                        nextindex.data,
                        length_so_far,
                        union_index.data,
                        array.length,
                    )
                )
                length_so_far += array.length
                nextcontents.extend(union_contents)

            elif isinstance(array, ak._v2.contents.emptyarray.EmptyArray):
                pass

            else:
                assert nexttags.nplike is self._nplike
                self._handle_error(
                    self._nplike[
                        "awkward_UnionArray_filltags_const",
                        nexttags.dtype.type,
                    ](
                        nexttags.data,
                        length_so_far,
                        array.length,
                        len(nextcontents),
                    )
                )

                assert nextindex.nplike is self._nplike
                self._handle_error(
                    self._nplike[
                        "awkward_UnionArray_fillindex_count", nextindex.dtype.type
                    ](nextindex.data, length_so_far, array.length)
                )

                length_so_far += array.length
                nextcontents.append(array)

        if len(nextcontents) > 127:
            raise ValueError("FIXME: handle UnionArray with more than 127 contents")

        next = ak._v2.contents.unionarray.UnionArray(
            nexttags, nextindex, nextcontents, None, parameters, self._nplike
        )

        # Given UnionArray's merging_strategy, tail is always empty, but just to be formal...

        if len(tail) == 0:
            return next

        reversed = tail[0]._reverse_merge(next)
        if len(tail) == 1:
            return reversed
        else:
            return reversed.mergemany(tail[1:])

    def fillna(self, value):
        contents = []
        for content in self._contents:
            contents.append(content.fillna(value))
        out = UnionArray(
            self._tags,
            self._index,
            contents,
            self._identifier,
            self._parameters,
            self._nplike,
        )
        return out.simplify_uniontype(True, False)

    def _localindex(self, axis, depth):
        posaxis = self.axis_wrap_if_negative(axis)
        if posaxis == depth:
            return self._localindex_axis0()
        else:
            contents = []
            for content in self._contents:
                contents.append(content._localindex(posaxis, depth))
            return UnionArray(
                self._tags,
                self._index,
                contents,
                self._identifier,
                self._parameters,
                self._nplike,
            )

    def _combinations(self, n, replacement, recordlookup, parameters, axis, depth):
        posaxis = self.axis_wrap_if_negative(axis)
        if posaxis == depth:
            return self._combinations_axis0(n, replacement, recordlookup, parameters)
        else:
            contents = []
            for content in self._contents:
                contents.append(
                    content._combinations(
                        n, replacement, recordlookup, parameters, posaxis, depth
                    )
                )
            return ak._v2.unionarray.UnionArray(
                self._tags,
                self._index,
                contents,
                self._identifier,
                self._parameters,
                self._nplike,
            )

    def numbers_to_type(self, name):
        contents = []
        for x in self._contents:
            contents.append(x.numbers_to_type(name))
        return ak._v2.contents.unionarray.UnionArray(
            self._tags,
            self._index,
            contents,
            self._identifier,
            self._parameters,
            self._nplike,
        )

    def _is_unique(self, negaxis, starts, parents, outlength):
        simplified = self.simplify_uniontype(True, True)
        if isinstance(simplified, ak._v2.contents.UnionArray):
            raise ValueError("cannot check if an irreducible UnionArray is unique")

        return simplified._is_unique(negaxis, starts, parents, outlength)

    def _unique(self, negaxis, starts, parents, outlength):
        simplified = self.simplify_uniontype(True, True)
        if isinstance(simplified, ak._v2.contents.UnionArray):
            raise ValueError("cannot make a unique irreducible UnionArray")

        return simplified._unique(negaxis, starts, parents, outlength)

    def _argsort_next(
        self,
        negaxis,
        starts,
        shifts,
        parents,
        outlength,
        ascending,
        stable,
        kind,
        order,
    ):
        if self.length == 0:
            return ak._v2.contents.NumpyArray(
                self._nplike.empty(0, np.int64), None, None, self._nplike
            )

        simplified = self.simplify_uniontype(mergebool=True)
        if simplified.length == 0:
            return ak._v2.contents.NumpyArray(
                self._nplike.empty(0, np.int64), None, None, self._nplike
            )

        if isinstance(simplified, ak._v2.contents.UnionArray):
            raise ValueError("cannot argsort an irreducible UnionArray")

        return simplified._argsort_next(
            negaxis, starts, shifts, parents, outlength, ascending, stable, kind, order
        )

    def _sort_next(
        self, negaxis, starts, parents, outlength, ascending, stable, kind, order
    ):
        if self.length == 0:
            return self

        simplified = self.simplify_uniontype(mergebool=True)
        if simplified.length == 0:
            return simplified

        if isinstance(simplified, ak._v2.contents.UnionArray):
            raise ValueError("cannot sort an irreducible UnionArray")

        return simplified._sort_next(
            negaxis, starts, parents, outlength, ascending, stable, kind, order
        )

    def _reduce_next(
        self,
        reducer,
        negaxis,
        starts,
        shifts,
        parents,
        outlength,
        mask,
        keepdims,
    ):
        simplified = self.simplify_uniontype(mergebool=True)
        if isinstance(simplified, UnionArray):
            raise ValueError(
                f"cannot call ak.{reducer.name} on an irreducible UnionArray"
            )

        return simplified._reduce_next(
            reducer,
            negaxis,
            starts,
            shifts,
            parents,
            outlength,
            mask,
            keepdims,
        )

    def _validityerror(self, path):
        for i in range(len(self.contents)):
            if isinstance(self.contents[i], ak._v2.contents.unionarray.UnionArray):
                return "{} contains {}, the operation that made it might have forgotten to call 'simplify_uniontype'".format(
                    type(self), type(self.contents[i])
                )
            if self._nplike.known_shape and self.index.length < self.tags.length:
                return f'at {path} ("{type(self)}"): len(index) < len(tags)'

            lencontents = self._nplike.empty(len(self.contents), dtype=np.int64)
            if self._nplike.known_shape:
                for i in range(len(self.contents)):
                    lencontents[i] = self.contents[i].length

            error = self._nplike[
                "awkward_UnionArray_validity",
                self.tags.dtype.type,
                self.index.dtype.type,
                np.int64,
            ](
                self.tags.data,
                self.index.data,
                self.tags.length,
                len(self.contents),
                lencontents,
            )

            if error.str is not None:
                if error.filename is None:
                    filename = ""
                else:
                    filename = " (in compiled code: " + error.filename.decode(
                        errors="surrogateescape"
                    ).lstrip("\n").lstrip("(")
                message = error.str.decode(errors="surrogateescape")
                return 'at {} ("{}"): {} at i={}{}'.format(
                    path, type(self), message, error.id, filename
                )

            for i in range(len(self.contents)):
                sub = self.contents[i].validityerror(path + f".content({i})")
                if sub != "":
                    return sub

            return ""

    def _nbytes_part(self):
        result = self.tags._nbytes_part() + self.index._nbytes_part()
        for content in self.contents:
            result = result + content._nbytes_part()
        if self.identifier is not None:
            result = result + self.identifier._nbytes_part()
        return result

    def _rpad(self, target, axis, depth, clip):
        posaxis = self.axis_wrap_if_negative(axis)
        if posaxis == depth:
            return self.rpad_axis0(target, clip)
        else:
            contents = []
            for content in self._contents:
                contents.append(content._rpad(target, posaxis, depth, clip))
            out = ak._v2.contents.unionarray.UnionArray(
                self.tags,
                self.index,
                contents,
                self._identifier,
                self._parameters,
                self._nplike,
            )
            return out.simplify_uniontype(True, False)

    def _to_arrow(self, pyarrow, mask_node, validbytes, length, options):
        nptags = self._tags.raw(numpy)
        npindex = self._index.raw(numpy)
        copied_index = False

        values = []
        for tag, content in enumerate(self._contents):
            selected_tags = nptags == tag
            this_index = npindex[selected_tags]

            # Arrow unions can't have masks; propagate validbytes down to the content.
            if validbytes is not None:
                # If this_index is a filtered permutation, we can just filter-permute
                # the mask to have the same order the content.
                if numpy.unique(this_index).shape[0] == this_index.shape[0]:
                    this_validbytes = numpy.zeros(this_index.shape[0], dtype=np.int8)
                    this_validbytes[this_index] = validbytes[selected_tags]

                # If this_index is not a filtered permutation, then we can't modify
                # the mask to fit the content. The same element in the content array
                # will appear multiple times in the union array, and it *might* be
                # presented as valid in some union array elements and masked in others.
                # The validbytes that recurses down to the next level can't have an
                # element that is both 0 (masked) and 1 (valid).
                else:
                    this_validbytes = validbytes[selected_tags]
                    content = content[this_index]
                    if not copied_index:
                        copied_index = True
                        npindex = numpy.array(npindex, copy=True)
                    npindex[selected_tags] = numpy.arange(
                        this_index.shape[0], dtype=npindex.dtype
                    )

            else:
                this_validbytes = None

            values.append(
                content._to_arrow(
                    pyarrow, mask_node, this_validbytes, this_index.shape[0], options
                )
            )

        types = pyarrow.union(
            [
                pyarrow.field(str(i), values[i].type).with_nullable(
                    mask_node is not None or self._contents[i].is_OptionType
                )
                for i in range(len(values))
            ],
            "dense",
            list(range(len(values))),
        )

        if not issubclass(npindex.dtype.type, np.int32):
            npindex = npindex.astype(np.int32)

        return pyarrow.Array.from_buffers(
            ak._v2._connect.pyarrow.to_awkwardarrow_type(
                types, options["extensionarray"], None, self
            ),
            nptags.shape[0],
            [
                None,
                ak._v2._connect.pyarrow.to_length(nptags, length),
                ak._v2._connect.pyarrow.to_length(npindex, length),
            ],
            children=values,
        )

    def _to_numpy(self, allow_missing):
        contents = [
            ak._v2.operations.convert.to_numpy(
                self.project(i), allow_missing=allow_missing
            )
            for i in range(len(self.contents))
        ]

        if any(isinstance(x, self._nplike.ma.MaskedArray) for x in contents):
            try:
                out = self._nplike.ma.concatenate(contents)
            except Exception:
                raise ValueError(f"cannot convert {self} into numpy.ma.MaskedArray")
        else:
            try:
                out = numpy.concatenate(contents)
            except Exception:
                raise ValueError(f"cannot convert {self} into np.ndarray")

        tags = numpy.asarray(self.tags)
        for tag, content in enumerate(contents):
            mask = tags == tag
            out[mask] = content
        return out

    def _completely_flatten(self, nplike, options):
        out = []
        for i in range(len(self._contents)):
            index = self._index[self._tags.data == i]
            out.extend(
                self._contents[i]
                ._carry(index, False, NestedIndexError)
                ._completely_flatten(nplike, options)
            )
        return out

    def _recursively_apply(
        self, action, depth, depth_context, lateral_context, options
    ):
        if options["return_array"]:

            def continuation():
                return UnionArray(
                    self._tags,
                    self._index,
                    [
                        content._recursively_apply(
                            action,
                            depth,
                            copy.copy(depth_context),
                            lateral_context,
                            options,
                        )
                        for content in self._contents
                    ],
                    self._identifier,
                    self._parameters if options["keep_parameters"] else None,
                    self._nplike,
                )

        else:

            def continuation():
                for content in self._contents:
                    content._recursively_apply(
                        action,
                        depth,
                        copy.copy(depth_context),
                        lateral_context,
                        options,
                    )

        result = action(
            self,
            depth=depth,
            depth_context=depth_context,
            lateral_context=lateral_context,
            continuation=continuation,
            options=options,
        )

        if isinstance(result, Content):
            return result
        elif result is None:
            return continuation()
        else:
            raise AssertionError(result)

    def packed(self):
        tags = self._tags.raw(self._nplike)
        original_index = index = self._index.raw(self._nplike)[: tags.shape[0]]

        contents = list(self._contents)

        for tag in range(len(self._contents)):
            is_tag = tags == tag
            num_tag = self._nplike.count_nonzero(is_tag)

            if len(contents[tag]) > num_tag:
                if original_index is index:
                    index = index.copy()
                index[is_tag] = self._nplike.arange(num_tag, dtype=index.dtype)
                contents[tag] = self.project(tag)

            contents[tag] = contents[tag].packed()

        return UnionArray(
            ak._v2.index.Index8(tags),
            ak._v2.index.Index(index),
            contents,
            self._identifier,
            self._parameters,
            self._nplike,
        )

    def _to_list(self, behavior):
        out = self._to_list_custom(behavior)
        if out is not None:
            return out

        tags = self._tags.raw(numpy)
        index = self._index.raw(numpy)
        contents = [x._to_list(behavior) for x in self._contents]

        out = [None] * tags.shape[0]
        for i, tag in enumerate(tags):
            out[i] = contents[tag][index[i]]
        return out

    def _to_nplike(self, nplike):
        index = self._index._to_nplike(nplike)
        contents = [content._to_nplike(nplike) for content in self._contents]
        return UnionArray(
            self._tags,
            index,
            contents,
            identifier=self.identifier,
            parameters=self.parameters,
            nplike=nplike,
        )

    def _to_json(
        self,
        nan_string,
        infinity_string,
        minus_infinity_string,
        complex_real_string,
        complex_imag_string,
    ):
        out = self._to_json_custom()
        if out is not None:
            return out

        tags = self._tags.raw(numpy)
        index = self._index.raw(numpy)
        contents = [
            x._to_json(
                nan_string,
                infinity_string,
                minus_infinity_string,
                complex_real_string,
                complex_imag_string,
            )
            for x in self._contents
        ]

        out = [None] * tags.shape[0]
        for i, tag in enumerate(tags):
            out[i] = contents[tag][index[i]]
        return out



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


def to_parquet(  # this will be going through Awkward-Dask
    array,
    where,
    explode_records=False,
    list_to32=False,
    string_to32=True,
    bytestring_to32=True,
    **options
):
    raise NotImplementedError


#     """
#     Args:
#         array: Data to write to a Parquet file.
#         where (str, Path, file-like object): Where to write the Parquet file.
#         explode_records (bool): If True, lists of records are written as
#             records of lists, so that nested fields become top-level fields
#             (which can be zipped when read back).
#         list_to32 (bool): If True, convert Awkward lists into 32-bit Arrow lists
#             if they're small enough, even if it means an extra conversion. Otherwise,
#             signed 32-bit #ak.layout.ListOffsetArray maps to Arrow `ListType` and
#             all others map to Arrow `LargeListType`.
#         string_to32 (bool): Same as the above for Arrow `string` and `large_string`.
#         bytestring_to32 (bool): Same as the above for Arrow `binary` and `large_binary`.
#         options: All other options are passed to pyarrow.parquet.ParquetWriter.
#             In particular, if no `schema` is given, a schema is derived from
#             the array type.

#     Writes an Awkward Array to a Parquet file (through pyarrow).

#         >>> array1 = ak.Array([[1, 2, 3], [], [4, 5], [], [], [6, 7, 8, 9]])
#         >>> ak.to_parquet(array1, "array1.parquet")

#     If the `array` does not contain records at top-level, the Arrow table will consist
#     of one field whose name is `""`.

#     Parquet files can maintain the distinction between "option-type but no elements are
#     missing" and "not option-type" at all levels, including the top level. However,
#     there is no distinction between `?union[X, Y, Z]]` type and `union[?X, ?Y, ?Z]` type.
#     Be aware of these type distinctions when passing data through Arrow or Parquet.

#     To make a partitioned Parquet dataset, use this function to write each Parquet
#     file to a directory (as separate invocations, probably in parallel with multiple
#     processes), then give them common metadata by calling `ak.to_parquet.dataset`.

#         >>> ak.to_parquet(array1, "directory-name/file1.parquet")
#         >>> ak.to_parquet(array2, "directory-name/file2.parquet")
#         >>> ak.to_parquet(array3, "directory-name/file3.parquet")
#         >>> ak.to_parquet.dataset("directory-name")

#     Then all of the flies in the collection can be addressed as one array. For example,

#         >>> dataset = ak.from_parquet("directory_name", lazy=True)

#     (If it is large, you will likely want to load it lazily.)

#     See also #ak.to_arrow, which is used as an intermediate step.
#     See also #ak.from_parquet.
#     """



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


# @ak._v2._connect.numpy.implements("argsort")
def argsort(array, axis=-1, ascending=True, stable=True, highlevel=True, behavior=None):

    """
    Args:
        array: Data for which to get a sorting index, possibly within nested
            lists.
        axis (int): The dimension at which this operation is applied. The
            outermost dimension is `0`, followed by `1`, etc., and negative
            values count backward from the innermost: `-1` is the innermost
            dimension, `-2` is the next level up, etc.
        ascending (bool): If True, the first value in each sorted group
            will be smallest, the last value largest; if False, the order
            is from largest to smallest.
        stable (bool): If True, use a stable sorting algorithm (introsort:
            a hybrid of quicksort, heapsort, and insertion sort); if False,
            use a sorting algorithm that is not guaranteed to be stable
            (heapsort).
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.layout.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.

    For example,

        >>> ak.argsort(ak.Array([[7.7, 5.5, 7.7], [], [2.2], [8.8, 2.2]]))
        <Array [[1, 0, 2], [], [0], [1, 0]] type='4 * var * int64'>

    The result of this function can be used to index other arrays with the
    same shape:

        >>> data = ak.Array([[7, 5, 7], [], [2], [8, 2]])
        >>> index = ak.argsort(index)
        >>> index
        <Array [[1, 0, 2], [], [0], [1, 0]] type='4 * var * int64'>
        >>> data[index]
        <Array [[5, 7, 7], [], [2], [2, 8]] type='4 * var * int64'>
    """
    layout = ak._v2.operations.convert.to_layout(
        array, allow_record=False, allow_other=False
    )
    out = layout.argsort(axis, ascending, stable)
    return ak._v2._util.wrap(out, behavior, highlevel)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE


import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_boolean():
    result = ak._v2.operations.convert.from_json_schema(
        " [ true ,false, true, true, false]    ",
        {"type": "array", "items": {"type": "boolean"}},
    )
    assert result.tolist() == [True, False, True, True, False]

    result = ak._v2.operations.convert.from_json_schema(
        "[]",
        {"type": "array", "items": {"type": "boolean"}},
    )
    assert result.tolist() == []


def test_integer():
    result = ak._v2.operations.convert.from_json_schema(
        " [ 1 ,2 ,3.0, 4, 5]  \n  ",
        {"type": "array", "items": {"type": "integer"}},
    )
    assert result.tolist() == [1, 2, 3, 4, 5]
    assert str(result.type) == "5 * int64"

    result = ak._v2.operations.convert.from_json_schema(
        "[ ]",
        {"type": "array", "items": {"type": "integer"}},
    )
    assert result.tolist() == []
    assert str(result.type) == "0 * int64"


def test_number():
    result = ak._v2.operations.convert.from_json_schema(
        " [ 1 ,2,3.14, 4, 5]",
        {"type": "array", "items": {"type": "number"}},
    )
    assert result.tolist() == [1, 2, 3.14, 4, 5]
    assert str(result.type) == "5 * float64"


def test_option_boolean():
    result = ak._v2.operations.convert.from_json_schema(
        " [ true ,false ,null , true, false]",
        {"type": "array", "items": {"type": ["boolean", "null"]}},
    )
    assert result.tolist() == [True, False, None, True, False]
    assert str(result.type) == "5 * ?bool"


def test_option_integer():
    result = ak._v2.operations.convert.from_json_schema(
        " [ 1 ,2,null,4, 5]",
        {"type": "array", "items": {"type": ["null", "integer"]}},
    )
    assert result.tolist() == [1, 2, None, 4, 5]
    assert str(result.type) == "5 * ?int64"

    result = ak._v2.operations.convert.from_json_schema(
        " [ ]",
        {"type": "array", "items": {"type": ["null", "integer"]}},
    )
    assert result.tolist() == []
    assert str(result.type) == "0 * ?int64"


def test_string():
    result = ak._v2.operations.convert.from_json_schema(
        r' [ "" ,"two","three \u2192 3", "\"four\"", "fi\nve"]',
        {"type": "array", "items": {"type": "string"}},
    )
    assert result.tolist() == ["", "two", "three \u2192 3", '"four"', "fi\nve"]

    result = ak._v2.operations.convert.from_json_schema(
        r"[]",
        {"type": "array", "items": {"type": "string"}},
    )
    assert result.tolist() == []


def test_option_string():
    result = ak._v2.operations.convert.from_json_schema(
        r' [ "" ,null ,"three \u2192 3", "\"four\"", "fi\nve"]',
        {"type": "array", "items": {"type": ["null", "string"]}},
    )
    assert result.tolist() == ["", None, "three \u2192 3", '"four"', "fi\nve"]

    result = ak._v2.operations.convert.from_json_schema(
        r"[]",
        {"type": "array", "items": {"type": ["null", "string"]}},
    )
    assert result.tolist() == []


def test_enum_string():
    result = ak._v2.operations.convert.from_json_schema(
        r'["three", "two", "one", "one", "two", "three"]',
        {"type": "array", "items": {"type": "string", "enum": ["one", "two", "three"]}},
    )
    assert result.tolist() == ["three", "two", "one", "one", "two", "three"]
    assert isinstance(result.layout, ak._v2.contents.IndexedArray)
    assert result.layout.index.data.tolist() == [2, 1, 0, 0, 1, 2]


def test_option_enum_string():
    result = ak._v2.operations.convert.from_json_schema(
        r'["three", "two", null, "one", "one", "two", "three"]',
        {
            "type": "array",
            "items": {"type": ["null", "string"], "enum": ["one", "two", "three"]},
        },
    )
    assert result.tolist() == ["three", "two", None, "one", "one", "two", "three"]
    assert isinstance(result.layout, ak._v2.contents.IndexedOptionArray)
    assert result.layout.index.data.tolist() == [2, 1, -1, 0, 0, 1, 2]


def test_array_integer():
    result = ak._v2.operations.convert.from_json_schema(
        " [ [ 1 ,2, 3], [], [4, 5]]",
        {"type": "array", "items": {"type": "array", "items": {"type": "integer"}}},
    )
    assert result.tolist() == [[1, 2, 3], [], [4, 5]]
    assert str(result.type) == "3 * var * int64"

    result = ak._v2.operations.convert.from_json_schema(
        "[]",
        {"type": "array", "items": {"type": "array", "items": {"type": "integer"}}},
    )
    assert result.tolist() == []


def test_regulararray_integer():
    result = ak._v2.operations.convert.from_json_schema(
        "[[1, 2, 3], [4, 5, 6]]",
        {
            "type": "array",
            "items": {
                "type": "array",
                "items": {"type": "integer"},
                "minItems": 3,
                "maxItems": 3,
            },
        },
    )
    assert result.tolist() == [[1, 2, 3], [4, 5, 6]]
    assert str(result.type) == "2 * 3 * int64"


def test_option_regulararray_integer():
    result = ak._v2.operations.convert.from_json_schema(
        "[[1, 2, 3], null, [4, 5, 6]]",
        {
            "type": "array",
            "items": {
                "type": ["array", "null"],
                "items": {"type": "integer"},
                "minItems": 3,
                "maxItems": 3,
            },
        },
    )
    assert result.tolist() == [[1, 2, 3], None, [4, 5, 6]]
    assert str(result.type) == "3 * option[3 * int64]"


def test_option_array_integer():
    result = ak._v2.operations.convert.from_json_schema(
        " [ [ 1 ,2,3 ],null,[ ], [4, 5]]",
        {
            "type": "array",
            "items": {"type": ["null", "array"], "items": {"type": "integer"}},
        },
    )
    assert result.tolist() == [[1, 2, 3], None, [], [4, 5]]
    assert str(result.type) == "4 * option[var * int64]"

    result = ak._v2.operations.convert.from_json_schema(
        "[]",
        {
            "type": "array",
            "items": {"type": ["null", "array"], "items": {"type": "integer"}},
        },
    )
    assert result.tolist() == []


def test_option_array_option_integer():
    result = ak._v2.operations.convert.from_json_schema(
        " [ [ 1 ,2,3 ],null,[ ] ,[null, 5]]",
        {
            "type": "array",
            "items": {
                "type": ["null", "array"],
                "items": {"type": ["integer", "null"]},
            },
        },
    )
    assert result.tolist() == [[1, 2, 3], None, [], [None, 5]]

    result = ak._v2.operations.convert.from_json_schema(
        "[]",
        {
            "type": "array",
            "items": {
                "type": ["null", "array"],
                "items": {"type": ["integer", "null"]},
            },
        },
    )
    assert result.tolist() == []


def test_array_array_integer():
    result = ak._v2.operations.convert.from_json_schema(
        " [ [ [ 1 ,2,3 ] ] ,[ [], [4, 5]], []]",
        {
            "type": "array",
            "items": {
                "type": "array",
                "items": {"type": "array", "items": {"type": "integer"}},
            },
        },
    )
    assert result.tolist() == [[[1, 2, 3]], [[], [4, 5]], []]


def test_record():
    result = ak._v2.operations.convert.from_json_schema(
        ' [ { "x" :1 ,"y":1.1},{"x": 2, "y": 2.2}, {"x": 3, "y": 3.3}]',
        {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {"x": {"type": "integer"}, "y": {"type": "number"}},
                "required": ["x", "y"],
            },
        },
    )
    assert result.tolist() == [
        {"x": 1, "y": 1.1},
        {"x": 2, "y": 2.2},
        {"x": 3, "y": 3.3},
    ]

    result = ak._v2.operations.convert.from_json_schema(
        "[]",
        {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {"x": {"type": "integer"}, "y": {"type": "number"}},
                "required": ["x", "y"],
            },
        },
    )
    assert result.tolist() == []


def test_option_record():
    result = ak._v2.operations.convert.from_json_schema(
        ' [ { "x" : 1 ,"y":1.1},null ,{"x": 2, "y": 2.2}, {"x": 3, "y": 3.3}]',
        {
            "type": "array",
            "items": {
                "type": ["object", "null"],
                "properties": {"x": {"type": "integer"}, "y": {"type": "number"}},
                "required": ["x", "y"],
            },
        },
    )
    assert result.tolist() == [
        {"x": 1, "y": 1.1},
        None,
        {"x": 2, "y": 2.2},
        {"x": 3, "y": 3.3},
    ]

    result = ak._v2.operations.convert.from_json_schema(
        "[]",
        {
            "type": "array",
            "items": {
                "type": ["object", "null"],
                "properties": {"x": {"type": "integer"}, "y": {"type": "number"}},
                "required": ["x", "y"],
            },
        },
    )
    assert result.tolist() == []


def test_top_record():
    result = ak._v2.operations.convert.from_json_schema(
        ' { "x" :1 ,"y":1.1}  ',
        {
            "type": "object",
            "properties": {"y": {"type": "number"}, "x": {"type": "integer"}},
            "required": ["x", "y"],
        },
    )
    assert result.tolist() == {"x": 1, "y": 1.1}



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


# @ak._v2._connect.numpy.implements("prod")
def prod(array, axis=None, keepdims=False, mask_identity=False, flatten_records=False):
    """
    Args:
        array: Array-like data (anything #ak.to_layout recognizes).
        axis (None or int): If None, combine all values from the array into
            a single scalar result; if an int, group by that axis: `0` is the
            outermost, `1` is the first level of nested lists, etc., and
            negative `axis` counts from the innermost: `-1` is the innermost,
            `-2` is the next level up, etc.
        keepdims (bool): If False, this reducer decreases the number of
            dimensions by 1; if True, the reduced values are wrapped in a new
            length-1 dimension so that the result of this operation may be
            broadcasted with the original array.
        mask_identity (bool): If True, reducing over empty lists results in
            None (an option type); otherwise, reducing over empty lists
            results in the operation's identity.
        flatten_records (bool): If True, axis=None combines fields from different
            records; otherwise, records raise an error.

    Multiplies elements of `array` (many types supported, including all
    Awkward Arrays and Records). The identity of multiplication is `1` and it
    is usually not masked. This operation is the same as NumPy's
    [prod](https://docs.scipy.org/doc/numpy/reference/generated/numpy.prod.html)
    if all lists at a given dimension have the same length and no None values,
    but it generalizes to cases where they do not.

    See #ak.sum for a more complete description of nested list and missing
    value (None) handling in reducers.
    """
    layout = ak._v2.operations.convert.to_layout(
        array, allow_record=False, allow_other=False
    )

    if axis is None:

        def reduce(xs):
            if len(xs) == 1:
                return xs[0]
            else:
                return layout.nplike.multiply(xs[0], reduce(xs[1:]))

        return reduce(
            [
                layout.nplike.prod(x)
                for x in layout.completely_flatten(
                    function_name="ak.prod", flatten_records=flatten_records
                )
            ]
        )

    else:
        behavior = ak._v2._util.behavior_of(array)
        out = layout.prod(axis=axis, mask=mask_identity, keepdims=keepdims)
        if isinstance(out, (ak._v2.contents.Content, ak._v2.record.Record)):
            return ak._v2._util.wrap(out, behavior)
        else:
            return out



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numbers

import numpy

import awkward.nplike
import awkward._util
import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


class NoError:
    str = None
    filename = None
    pass_through = False
    attempt = ak._util.kSliceNone
    id = ak._util.kSliceNone


class NoKernel:
    def __call__(self, *args):
        return NoError()


class UnknownLengthType:
    def __repr__(self):
        return "UnknownLength"

    def __str__(self):
        return "??"

    def __eq__(self, other):
        return isinstance(other, UnknownLengthType)

    def __add__(self, other):
        return UnknownLength

    def __radd__(self, other):
        return UnknownLength

    def __sub__(self, other):
        return UnknownLength

    def __rsub__(self, other):
        return UnknownLength

    def __mul__(self, other):
        return UnknownLength

    def __rmul__(self, other):
        return UnknownLength

    def __truediv__(self, other):
        return UnknownLength

    def __floordiv__(self, other):
        return UnknownLength

    def __rdiv__(self, other):
        return UnknownLength

    def __rfloordiv__(self, other):
        return UnknownLength

    def __lt__(self, other):
        return False

    def __le__(self, other):
        return False

    def __gt__(self, other):
        return False

    def __ge__(self, other):
        return False


UnknownLength = UnknownLengthType()


def _emptyarray(x):
    if isinstance(x, UnknownScalar):
        return numpy.empty(0, x._dtype)
    elif hasattr(x, "dtype"):
        return numpy.empty(0, x.dtype)
    else:
        return numpy.empty(0, numpy.array(x).dtype)


class UnknownScalar:
    def __init__(self, dtype):
        self._dtype = dtype

    @property
    def dtype(self):
        return self._dtype

    def __repr__(self):
        return f"UnknownScalar({self._dtype!r})"

    def __str__(self):
        return f"unknown-{str(self._dtype)}"

    def __eq__(self, other):
        return isinstance(other, UnknownScalar) and self._dtype == other._dtype

    def __add__(self, other):
        return UnknownScalar((_emptyarray(self) + _emptyarray(other)).dtype)

    def __radd__(self, other):
        return UnknownScalar((_emptyarray(self) + _emptyarray(other)).dtype)

    def __sub__(self, other):
        return UnknownScalar((_emptyarray(self) - _emptyarray(other)).dtype)

    def __rsub__(self, other):
        return UnknownScalar((_emptyarray(self) - _emptyarray(other)).dtype)

    def __mul__(self, other):
        return UnknownScalar((_emptyarray(self) * _emptyarray(other)).dtype)

    def __rmul__(self, other):
        return UnknownScalar((_emptyarray(self) * _emptyarray(other)).dtype)

    def __truediv__(self, other):
        return UnknownScalar((_emptyarray(self) / _emptyarray(other)).dtype)

    def __floordiv__(self, other):
        return UnknownScalar((_emptyarray(self) // _emptyarray(other)).dtype)

    def __lt__(self, other):
        return False

    def __le__(self, other):
        return False

    def __gt__(self, other):
        return False

    def __ge__(self, other):
        return False


class MaybeNone:
    def __init__(self, content):
        self._content = content

    @property
    def content(self):
        return self._content

    def __eq__(self, other):
        if isinstance(other, MaybeNone):
            return self._content == other._content
        else:
            return False

    def __repr__(self):
        return f"MaybeNone({self._content!r})"


class OneOf:
    def __init__(self, contents):
        self._contents = contents

    @property
    def contents(self):
        return self._contents

    def __eq__(self, other):
        if isinstance(other, OneOf):
            return set(self._contents) == set(other._contents)
        else:
            return False

    def __repr__(self):
        return f"OneOf({self._contents!r})"


def _length_after_slice(slice, original_length):
    start, stop, step = slice.indices(original_length)
    assert step != 0

    if (step > 0 and stop - start > 0) or (step < 0 and stop - start < 0):
        d, m = divmod(abs(start - stop), abs(step))
        return d + (1 if m != 0 else 0)
    else:
        return 0


class TypeTracerArray:
    @classmethod
    def from_array(cls, array, dtype=None):
        if isinstance(array, ak._v2.index.Index):
            array = array.data
        if dtype is None:
            dtype = array.dtype
        return cls(dtype, shape=array.shape)

    def __init__(self, dtype, shape=None):
        self._dtype = np.dtype(dtype)
        self.shape = shape

    def __repr__(self):
        dtype = repr(self._dtype)
        shape = ""
        if self._shape != (UnknownLength,):
            shape = ", " + repr(self._shape)
        return f"TypeTracerArray({dtype}{shape})"

    @property
    def dtype(self):
        return self._dtype

    @property
    def shape(self):
        return self._shape

    @shape.setter
    def shape(self, value):
        if ak._v2._util.isint(value):
            value = (value,)
        elif value is None or isinstance(value, (UnknownLengthType, UnknownScalar)):
            value = (UnknownLength,)
        elif not isinstance(value, tuple):
            value = tuple(value)
        self._shape = value

    @property
    def strides(self):
        out = (self._dtype.itemsize,)
        for x in self._shape[:0:-1]:
            out = (x * out[0],) + out
        return out

    @property
    def nplike(self):
        return TypeTracer.instance()

    @property
    def ndim(self):
        return len(self._shape)

    def forget_length(self):
        return type(self)(self._dtype, (UnknownLength,) + self._shape[1:])

    def __iter__(self):
        raise AssertionError(
            "bug in Awkward Array: attempt to convert TypeTracerArray into a concrete array"
        )

    def __array__(self, *args, **kwargs):
        raise AssertionError(
            "bug in Awkward Array: attempt to convert TypeTracerArray into a concrete array"
        )

    def itemsize(self):
        return self._dtype.itemsize

    class _CTypes:
        data = 0

    @property
    def ctypes(self):
        return self._CTypes

    def __len__(self):
        raise AssertionError(
            "bug in Awkward Array: attempt to get length of a TypeTracerArray"
        )

    def __setitem__(self, where, what):
        raise AssertionError(
            "bug in Awkward Array: attempt to set values of a TypeTracerArray"
        )

    def __getitem__(self, where):
        if isinstance(where, tuple):
            try:
                i = where.index(Ellipsis)
            except ValueError:
                pass
            else:
                before, after = where[:i], where[i + 1 :]
                missing = max(0, len(self._shape) - (len(before) + len(after)))
                where = before + (slice(None, None, None),) * missing + after

        if ak._v2._util.isint(where):
            if len(self._shape) == 1:
                if where == 0:
                    return UnknownScalar(self._dtype)
                else:
                    return UnknownScalar(self._dtype)
            else:
                return TypeTracerArray(self._dtype, self._shape[1:])

        elif isinstance(where, slice):
            return TypeTracerArray(self._dtype, (UnknownLength,) + self._shape[1:])

        elif (
            hasattr(where, "dtype")
            and hasattr(where, "shape")
            and issubclass(where.dtype.type, np.integer)
        ):
            assert len(self._shape) != 0
            return TypeTracerArray(self._dtype, where.shape + self._shape[1:])

        elif (
            hasattr(where, "dtype")
            and hasattr(where, "shape")
            and issubclass(where.dtype.type, (np.bool_, bool))
        ):
            assert len(self._shape) != 0
            return TypeTracerArray(self._dtype, (UnknownLength,) + self._shape[1:])

        elif isinstance(where, tuple) and any(
            hasattr(x, "dtype") and hasattr(x, "shape") for x in where
        ):
            for num_basic, wh in enumerate(where):  # noqa: B007
                if not isinstance(wh, slice):
                    break

            if num_basic != 0:
                tmp = self.__getitem__(where[:num_basic])
                basic_shape = tmp._shape[:num_basic]
            else:
                basic_shape = ()

            shapes = []
            for j in range(num_basic, len(where)):
                wh = where[j]
                if ak._v2._util.isint(wh):
                    shapes.append(numpy.array(0))
                elif hasattr(wh, "dtype") and hasattr(wh, "shape"):
                    sh = [
                        1 if isinstance(x, UnknownLengthType) else int(x)
                        for x in wh.shape
                    ]
                    shapes.append(
                        numpy.lib.stride_tricks.as_strided(
                            numpy.array(0), shape=sh, strides=[0] * len(sh)
                        )
                    )
                else:
                    raise NotImplementedError(repr(wh))

            slicer_shape = numpy.broadcast_arrays(*shapes)[0].shape

            shape = basic_shape + slicer_shape + self._shape[num_basic + len(shapes) :]
            assert len(shape) != 0

            return TypeTracerArray(self._dtype, (UnknownLength,) + shape[1:])

        elif (
            isinstance(where, tuple)
            and len(where) > 0
            and (ak._v2._util.isint(where[0]) or isinstance(where[0], slice))
        ):
            head, tail = where[0], where[1:]
            next = self.__getitem__(head)

            inner_shape = next.shape[1:]
            after_shape = []
            for i, wh in enumerate(tail):
                if isinstance(wh, int):
                    pass
                elif isinstance(wh, slice):
                    after_shape.append(_length_after_slice(wh, inner_shape[i]))
                else:
                    raise NotImplementedError(repr(wh))

            shape = (next._shape[0],) + tuple(after_shape)
            return TypeTracerArray(self._dtype, shape)

        else:
            raise NotImplementedError(repr(where))

    def __lt__(self, other):
        if isinstance(other, numbers.Real):
            return TypeTracerArray(np.bool_, self._shape)
        else:
            return NotImplemented

    def __le__(self, other):
        if isinstance(other, numbers.Real):
            return TypeTracerArray(np.bool_, self._shape)
        else:
            return NotImplemented

    def __gt__(self, other):
        if isinstance(other, numbers.Real):
            return TypeTracerArray(np.bool_, self._shape)
        else:
            return NotImplemented

    def __ge__(self, other):
        if isinstance(other, numbers.Real):
            return TypeTracerArray(np.bool_, self._shape)
        else:
            return NotImplemented

    def reshape(self, *args):
        if len(args) == 1 and isinstance(args[0], tuple):
            args = args[0]

        assert len(args) != 0
        assert ak._v2._util.isint(args[0]) or isinstance(args[0], UnknownLengthType)
        assert all(ak._v2._util.isint(x) for x in args[1:])
        assert all(x >= 0 for x in args[1:])

        return TypeTracerArray(self._dtype, (UnknownLength,) + args[1:])

    def copy(self):
        return self


unset = object()


class TypeTracer(ak.nplike.NumpyLike):
    known_data = False
    known_shape = False

    def to_rectilinear(self, array, *args, **kwargs):
        raise NotImplementedError

    def __getitem__(self, name_and_types):
        return NoKernel()

    @property
    def ma(self):
        raise NotImplementedError

    @property
    def char(self):
        raise NotImplementedError

    @property
    def ndarray(self):
        return TypeTracerArray

    def raw(self, array, nplike):
        assert isinstance(array.nplike, TypeTracer)

        if isinstance(nplike, TypeTracer):
            return TypeTracerArray.from_array(array)
        elif isinstance(array, TypeTracerArray):
            return self
        elif hasattr(nplike, "known_data") and nplike.known_data:
            raise TypeError(
                "Converting a TypeTracer nplike to a nplike with `known_data=True` is not possible"
            )
        else:
            raise TypeError(
                "Invalid nplike, choose between nplike.Numpy, nplike.Cupy, Typetracer"
            )

    ############################ array creation

    def array(self, data, dtype=unset, **kwargs):
        # data[, dtype=[, copy=]]
        if dtype is unset:
            dtype = data.dtype
        return TypeTracerArray.from_array(data, dtype=dtype)

    def asarray(self, array, dtype=unset, **kwargs):
        # array[, dtype=][, order=]
        if dtype is unset:
            dtype = array.dtype
        return TypeTracerArray.from_array(array, dtype=dtype)

    def ascontiguousarray(self, array, dtype=unset, **kwargs):
        # array[, dtype=]
        if dtype is unset:
            dtype = array.dtype
        return TypeTracerArray.from_array(array, dtype=dtype)

    def isscalar(self, *args, **kwargs):
        raise NotImplementedError

    def frombuffer(self, *args, **kwargs):
        # array[, dtype=]
        raise NotImplementedError

    def zeros(self, shape, dtype=np.float64, **kwargs):
        # shape/len[, dtype=]
        return TypeTracerArray(dtype, shape)

    def ones(self, shape, dtype=np.float64, **kwargs):
        # shape/len[, dtype=]
        return TypeTracerArray(dtype, shape)

    def empty(self, shape, dtype=np.float64, **kwargs):
        # shape/len[, dtype=]
        return TypeTracerArray(dtype, shape)

    def full(self, shape, value, dtype=unset, **kwargs):
        # shape/len, value[, dtype=]
        if dtype is unset:
            dtype = numpy.array(value).dtype
        return TypeTracerArray(dtype, shape)

    def zeros_like(self, *args, **kwargs):
        # array
        raise NotImplementedError

    def ones_like(self, *args, **kwargs):
        # array
        raise NotImplementedError

    def full_like(self, *args, **kwargs):
        # array, fill_value
        raise NotImplementedError

    def arange(self, *args, **kwargs):
        # stop[, dtype=]
        # start, stop[, dtype=]
        # start, stop, step[, dtype=]
        assert 1 <= len(args) <= 3
        assert (
            "dtype" in kwargs
        ), "internal error: calling arange without dtype (platform dependence)"

        if len(args) == 1:
            start, stop, step = 0, args[0], 1
        elif len(args) == 2:
            start, stop, step = args[0], args[1], 1
        elif len(args) == 3:
            start, stop, step = args[0], args[1], args[2]

        if (
            ak._v2._util.isint(start)
            and ak._v2._util.isint(stop)
            and ak._v2._util.isint(step)
        ):
            length = max(0, (stop - start + (step - (1 if step > 0 else -1))) // step)

        return TypeTracerArray(kwargs["dtype"], (length,))

    def meshgrid(self, *args, **kwargs):
        # *arrays, indexing="ij"
        raise NotImplementedError

    ############################ testing

    def shape(self, *args, **kwargs):
        # array
        raise NotImplementedError

    def array_equal(self, *args, **kwargs):
        # array1, array2
        return False

    def size(self, *args, **kwargs):
        # array
        raise NotImplementedError

    def searchsorted(self, *args, **kwargs):
        # haystack, needle, side="right"
        raise NotImplementedError

    def argsort(self, array, *args, **kwargs):
        # array
        return TypeTracerArray(np.int64, array.shape)

    ############################ manipulation

    def broadcast_arrays(self, *arrays):
        # array1[, array2[, ...]]

        if len(arrays) == 0:
            return []

        next = []
        maxdim = 0
        for x in arrays:
            if not hasattr(x, "shape"):
                next.append(numpy.array(x))
            else:
                next.append(x)
                maxdim = max(maxdim, len(x.shape))

        if maxdim == 0:
            return next

        first, *rest = next
        shape = list(first.shape[1:])
        for x in rest:
            thisshape = x.shape[1:]
            if len(shape) < len(thisshape):
                shape = [1] * (len(thisshape) - len(shape)) + shape
            elif len(shape) > len(thisshape):
                thisshape = (1,) * (len(shape) - len(thisshape)) + thisshape
            for i in range(len(shape)):  # pylint: disable=consider-using-enumerate
                if shape[i] == 1 and thisshape[i] != 1:
                    shape[i] = thisshape[i]
                elif shape[i] != 1 and thisshape[i] != 1 and shape[i] != thisshape[i]:
                    raise ValueError(
                        "shape mismatch: objects cannot be broadcast to a single shape"
                    )

        return [
            TypeTracerArray(x.dtype, [UnknownLength] + shape) for x in [first] + rest
        ]

    def add(self, *args, **kwargs):
        # array1, array2[, out=]
        raise NotImplementedError

    def cumsum(self, *args, **kwargs):
        # arrays[, out=]
        raise NotImplementedError

    def cumprod(self, *args, **kwargs):
        # arrays[, out=]
        raise NotImplementedError

    def nonzero(self, array):
        # array
        return (TypeTracerArray(np.int64, (UnknownLength,)),) * len(array.shape)

    def unique(self, *args, **kwargs):
        # array
        raise NotImplementedError

    def concatenate(self, arrays):
        inner_shape = None
        emptyarrays = []
        for x in arrays:
            if inner_shape is None:
                inner_shape = x.shape[1:]
            elif inner_shape != x.shape[1:]:
                raise ValueError(
                    "inner dimensions don't match in concatenate: {} vs {}".format(
                        inner_shape, x.shape[1:]
                    )
                )
            emptyarrays.append(_emptyarray(x))

        if inner_shape is None:
            raise ValueError("need at least one array to concatenate")

        return TypeTracerArray(
            numpy.concatenate(emptyarrays).dtype, (UnknownLength,) + inner_shape
        )

    def repeat(self, *args, **kwargs):
        # array, int
        # array1, array2
        raise NotImplementedError

    def stack(self, *args, **kwargs):
        # arrays
        raise NotImplementedError

    def vstack(self, *args, **kwargs):
        # arrays
        raise NotImplementedError

    def packbits(self, *args, **kwargs):
        # array
        raise NotImplementedError

    def unpackbits(self, *args, **kwargs):
        # array
        raise NotImplementedError

    def atleast_1d(self, *args, **kwargs):
        # *arrays
        raise NotImplementedError

    def broadcast_to(self, *args, **kwargs):
        # array, shape
        raise NotImplementedError

    def append(self, *args, **kwargs):
        # array, element
        raise NotImplementedError

    def where(self, *args, **kwargs):
        # array, element
        raise NotImplementedError

    ############################ ufuncs

    def sqrt(self, *args, **kwargs):
        # array
        raise NotImplementedError

    def exp(self, *args, **kwargs):
        # array
        raise NotImplementedError

    def true_divide(self, *args, **kwargs):
        # array1, array2
        raise NotImplementedError

    def bitwise_or(self, *args, **kwargs):
        # array1, array2[, out=output]
        raise NotImplementedError

    def logical_and(self, *args, **kwargs):
        # array1, array2
        raise NotImplementedError

    def equal(self, *args, **kwargs):
        # array1, array2
        raise NotImplementedError

    def ceil(self, *args, **kwargs):
        # array
        raise NotImplementedError

    ############################ almost-ufuncs

    def nan_to_num(self, *args, **kwargs):
        # array, copy=True, nan=0.0, posinf=None, neginf=None
        raise NotImplementedError

    def isclose(self, *args, **kwargs):
        # a, b, rtol=1e-05, atol=1e-08, equal_nan=False
        raise NotImplementedError

    ############################ reducers

    def all(self, array, prefer):
        # array
        return prefer

    def any(self, array, prefer):
        # array
        return prefer

    def count_nonzero(self, *args, **kwargs):
        # array
        raise NotImplementedError

    def sum(self, *args, **kwargs):
        # array
        raise NotImplementedError

    def prod(self, *args, **kwargs):
        # array
        raise NotImplementedError

    def min(self, *args, **kwargs):
        # array
        raise NotImplementedError

    def max(self, *args, **kwargs):
        # array
        raise NotImplementedError

    def argmin(self, *args, **kwargs):
        # array[, axis=]
        raise NotImplementedError

    def argmax(self, *args, **kwargs):
        # array[, axis=]
        raise NotImplementedError

    def array_str(
        self, array, max_line_width=unset, precision=unset, suppress_small=unset
    ):
        # array, max_line_width, precision=None, suppress_small=None
        return "[?? ... ??]"

    def datetime_as_string(self, *args, **kwargs):
        raise NotImplementedError



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numbers

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


def type(array):

    """
    The high-level type of an `array` (many types supported, including all
    Awkward Arrays and Records) as #ak.types.Type objects.

    The high-level type ignores #layout differences like
    #ak.layout.ListArray64 versus #ak.layout.ListOffsetArray64, but
    not differences like "regular-sized lists" (i.e.
    #ak.layout.RegularArray) versus "variable-sized lists" (i.e.
    #ak.layout.ListArray64 and similar).

    Types are rendered as [Datashape](https://datashape.readthedocs.io/)
    strings, which makes the same distinctions.

    For example,

        ak.Array([[{"x": 1.1, "y": [1]}, {"x": 2.2, "y": [2, 2]}],
                  [],
                  [{"x": 3.3, "y": [3, 3, 3]}]])

    has type

        3 * var * {"x": float64, "y": var * int64}

    but

        ak.Array(np.arange(2*3*5).reshape(2, 3, 5))

    has type

        2 * 3 * 5 * int64

    Some cases, like heterogeneous data, require [extensions beyond the
    Datashape specification](https://github.com/blaze/datashape/issues/237).
    For example,

        ak.Array([1, "two", [3, 3, 3]])

    has type

        3 * union[int64, string, var * int64]

    but "union" is not a Datashape type-constructor. (Its syntax is
    similar to existing type-constructors, so it's a plausible addition
    to the language.)
    """
    if array is None:
        return ak._v2.types.UnknownType()

    elif isinstance(
        array,
        tuple(x.type for x in ak._v2.types.numpytype._dtype_to_primitive_dict),
    ):
        return ak._v2.types.NumpyType(
            ak._v2.types.numpytype._dtype_to_primitive_dict[array.dtype]
        )

    elif isinstance(array, (bool, np.bool_)):
        return ak._v2.types.NumpyType("bool")

    elif isinstance(array, numbers.Integral):
        return ak._v2.types.NumpyType("int64")

    elif isinstance(array, numbers.Real):
        return ak._v2.types.NumpyType("float64")

    elif isinstance(
        array,
        (
            ak._v2.highlevel.Array,
            ak._v2.highlevel.Record,
            ak._v2.highlevel.ArrayBuilder,
        ),
    ):
        return array.type

    elif isinstance(array, np.ndarray):
        if len(array.shape) == 0:
            return type(array.reshape((1,))[0])
        else:
            try:
                out = ak._v2.types.numpytype._dtype_to_primitive_dict[array.dtype.type]
            except KeyError:
                raise TypeError(
                    "numpy array type is unrecognized by awkward: %r" % array.dtype.type
                )
            out = ak._v2.types.NumpyType(out)
            for x in array.shape[-1:0:-1]:
                out = ak._v2.types.RegularType(out, x)
            return ak._v2.types.ArrayType(out, array.shape[0])

    elif isinstance(array, ak.layout.ArrayBuilder):
        return NotImplementedError

    elif isinstance(array, (ak._v2.contents.Content, ak._v2.record.Record)):
        return array.form.type

    elif isinstance(
        array,
        (
            ak.highlevel.Array,
            ak.highlevel.Record,
            ak.highlevel.ArrayBuilder,
            ak.layout.Content,
            ak.layout.Record,
        ),
    ):
        raise TypeError("do not use ak._v2.operations.describe.type on v1 arrays")

    else:
        raise TypeError(f"unrecognized array type: {array!r}")



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


def packed(array, highlevel=True, behavior=None):
    """
    Args:
        array: Array whose internal structure will be packed.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.layout.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.

    Returns an array with the same type and values as the input, but with packed inner structures:

    - #ak.layout.NumpyArray becomes C-contiguous (if it isn't already)
    - #ak.layout.RegularArray trims unreachable content
    - #ak.layout.ListArray becomes #ak.layout.ListOffsetArray, making all list data contiguous
    - #ak.layout.ListOffsetArray starts at `offsets[0] == 0`, trimming unreachable content
    - #ak.layout.RecordArray trims unreachable contents
    - #ak.layout.IndexedArray gets projected
    - #ak.layout.IndexedOptionArray remains an #ak.layout.IndexedOptionArray (with simplified `index`) if it contains records, becomes #ak.layout.ByteMaskedArray otherwise
    - #ak.layout.ByteMaskedArray becomes an #ak.layout.IndexedOptionArray if it contains records, stays a #ak.layout.ByteMaskedArray otherwise
    - #ak.layout.BitMaskedArray becomes an #ak.layout.IndexedOptionArray if it contains records, stays a #ak.layout.BitMaskedArray otherwise
    - #ak.layout.UnionArray gets projected contents
    - #ak.layout.Record becomes a record over a single-item #ak.layout.RecordArray

    Example:

        >>> a = ak.Array([[1, 2, 3], [], [4, 5], [6], [7, 8, 9, 10]])
        >>> b = a[::-1]
        >>> b
        <Array [[7, 8, 9, 10], [6, ... [], [1, 2, 3]] type='5 * var * int64'>
        >>> b.layout
        <ListArray64>
            <starts><Index64 i="[6 5 3 3 0]" offset="0" length="5" at="0x55e091c2b1f0"/></starts>
            <stops><Index64 i="[10 6 5 3 3]" offset="0" length="5" at="0x55e091a6ce80"/></stops>
            <content><NumpyArray format="l" shape="10" data="1 2 3 4 5 6 7 8 9 10" at="0x55e091c47260"/></content>
        </ListArray64>
        >>> c = ak.packed(b)
        >>> c
        <Array [[7, 8, 9, 10], [6, ... [], [1, 2, 3]] type='5 * var * int64'>
        >>> c.layout
        <ListOffsetArray64>
            <offsets><Index64 i="[0 4 5 7 7 10]" offset="0" length="6" at="0x55e091b077a0"/></offsets>
            <content><NumpyArray format="l" shape="10" data="7 8 9 10 6 4 5 1 2 3" at="0x55e091d04d30"/></content>
        </ListOffsetArray64>

    Performing these operations will minimize the output size of data sent to
    #ak.to_buffers (though conversions through Arrow, #ak.to_arrow and
    #ak.to_parquet, do not need this because packing is part of that conversion).

    See also #ak.to_buffers.
    """
    layout = ak._v2.operations.convert.to_layout(
        array, allow_record=True, allow_other=False
    )
    out = layout.packed()
    return ak._v2._util.wrap(out, behavior, highlevel)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


# @ak._v2._connect.numpy.implements("where")
def where(condition, *args, **kwargs):

    """
    Args:
        condition (np.ndarray or rectilinear #ak.Array of booleans): In the
            three-argument form of this function (`condition`, `x`, `y`),
            True values in `condition` select values from `x` and False
            values in `condition` select values from `y`.
        x: Data with the same length as `condition`.
        y: Data with the same length as `condition`.
        mergebool (bool, default is True): If True, boolean and nummeric data
            can be combined into the same buffer, losing information about
            False vs `0` and True vs `1`; otherwise, they are kept in separate
            buffers with distinct types (using an #ak.layout.UnionArray8_64).
        highlevel (bool, default is True): If True, return an #ak.Array;
            otherwise, return a low-level #ak.layout.Content subclass.

    This function has a one-argument form, `condition` without `x` or `y`, and
    a three-argument form, `condition`, `x`, and `y`. In the one-argument form,
    it is completely equivalent to NumPy's
    [nonzero](https://docs.scipy.org/doc/numpy/reference/generated/numpy.nonzero.html)
    function.

    In the three-argument form, it acts as a vectorized ternary operator:
    `condition`, `x`, and `y` must all have the same length and

        output[i] = x[i] if condition[i] else y[i]

    for all `i`. The structure of `x` and `y` do not need to be the same; if
    they are incompatible types, the output will have #ak.type.UnionType.
    """
    mergebool, highlevel = ak._v2._util.extra(
        (), kwargs, [("mergebool", True), ("highlevel", True)]
    )

    akcondition = ak._v2.operations.convert.to_layout(
        condition, allow_record=False, allow_other=False
    )

    if len(args) == 0:
        nplike = ak.nplike.of(akcondition)

        akcondition = ak._v2.contents.NumpyArray(
            ak._v2.operations.convert.to_numpy(akcondition)
        )
        out = nplike.nonzero(ak._v2.operations.convert.to_numpy(akcondition))
        if highlevel:
            return tuple(
                ak._v2._util.wrap(
                    ak._v2.contents.NumpyArray(x), ak._v2._util.behavior_of(condition)
                )
                for x in out
            )
        else:
            return tuple(ak._v2.contents.NumpyArray(x) for x in out)

    elif len(args) == 1:
        raise ValueError("either both or neither of x and y should be given")

    elif len(args) == 2:
        left, right = (
            ak._v2.operations.convert.to_layout(x, allow_record=False, allow_other=True)
            for x in args
        )
        good_arrays = [akcondition]
        if isinstance(left, ak._v2.contents.Content):
            good_arrays.append(left)
        if isinstance(right, ak._v2.contents.Content):
            good_arrays.append(right)
        nplike = ak.nplike.of(*good_arrays)

        def action(inputs, **kwargs):
            akcondition, left, right = inputs
            if isinstance(akcondition, ak._v2.contents.NumpyArray):
                npcondition = nplike.asarray(akcondition)
                tags = ak._v2.index.Index8((npcondition == 0).view(np.int8))
                index = ak._v2.index.Index64(nplike.arange(len(tags), dtype=np.int64))
                if not isinstance(left, ak._v2.contents.Content):
                    left = ak._v2.contents.NumpyArray(nplike.repeat(left, len(tags)))
                if not isinstance(right, ak._v2.contents.Content):
                    right = ak._v2.contents.NumpyArray(nplike.repeat(right, len(tags)))
                tmp = ak._v2.contents.UnionArray(tags, index, [left, right])
                return (tmp.simplify_uniontype(mergebool=mergebool),)
            else:
                return None

        behavior = ak._v2._util.behavior_of(condition, *args)
        out = ak._v2._broadcasting.broadcast_and_apply(
            [akcondition, left, right],
            action,
            behavior,
            numpy_to_regular=True,
        )

        return ak._v2._util.wrap(out[0], behavior, highlevel)

    else:
        raise TypeError(
            "where() takes from 1 to 3 positional arguments but {} were "
            "given".format(len(args) + 1)
        )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE


import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    array = ak.from_numpy(np.zeros((3, 3, 5)))
    flattened = ak.flatten(array, axis=None)
    assert flattened.ndim == 1



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    index_of_index = ak.Array(
        ak.layout.IndexedOptionArray64(
            ak.layout.Index64(np.r_[0, 1, 2, -1]),
            ak.layout.IndexedOptionArray64(
                ak.layout.Index64(np.r_[0, -1, 2, 3]),
                ak.layout.NumpyArray(np.r_[1, 2, 3, 4]),
            ),
        )
    )

    mask = ak.is_none(index_of_index)
    assert ak.to_list(mask) == [False, True, False, True]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

# First, transition all the _v2 code to start using implementations in this file.
# Then build up the high-level replacements.

# import re
# import os.path
# import warnings
import setuptools
import os
import numbers

from collections.abc import Mapping

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()

win = os.name == "nt"
bits32 = ak.nplike.numpy.iinfo(np.intp).bits == 32

# matches include/awkward/common.h
kMaxInt8 = 127  # 2**7  - 1
kMaxUInt8 = 255  # 2**8  - 1
kMaxInt32 = 2147483647  # 2**31 - 1
kMaxUInt32 = 4294967295  # 2**32 - 1
kMaxInt64 = 9223372036854775806  # 2**63 - 2: see below
kSliceNone = kMaxInt64 + 1  # for Slice::none()
kMaxLevels = 48

_backends = {
    "cpu": ak.nplike.Numpy,
    "cuda": ak.nplike.Cupy,
}


def regularize_backend(backend):
    if backend in _backends:
        return _backends[backend].instance()
    else:
        raise ValueError("The available backends for now are `cpu` and `cuda`.")


def parse_version(version):
    return setuptools.extern.packaging.version.parse(version)


def numpy_at_least(version):
    import numpy

    return parse_version(numpy.__version__) >= parse_version(version)


def in_module(obj, modulename):
    m = type(obj).__module__
    return m == modulename or m.startswith(modulename + ".")


def is_file_path(x):
    try:
        return os.path.isfile(x)
    except ValueError:
        return False


def isint(x):
    return isinstance(x, (int, numbers.Integral, np.integer)) and not isinstance(
        x, (bool, np.bool_)
    )


def isnum(x):
    return isinstance(x, (int, float, numbers.Real, np.number)) and not isinstance(
        x, (bool, np.bool_)
    )


def isstr(x):
    return isinstance(x, str)


def tobytes(array):
    if hasattr(array, "tobytes"):
        return array.tobytes()
    else:
        return array.tostring()


def little_endian(array):
    return array.astype(array.dtype.newbyteorder("<"), copy=False)


###############################################################################

# # Enable warnings for the Awkward package
# warnings.filterwarnings("default", module="awkward.*")


# def deprecate(
#     message,
#     version,
#     date=None,
#     will_be="an error",
#     category=DeprecationWarning,
#     stacklevel=2,
# ):
#     if date is None:
#         date = ""
#     else:
#         date = " (target date: " + date + ")"
#     warning = """In version {0}{1}, this will be {2}.

# To raise these warnings as errors (and get stack traces to find out where they're called), run

#     import warnings
#     warnings.filterwarnings("error", module="awkward.*")

# after the first `import awkward` or use `@pytest.mark.filterwarnings("error:::awkward.*")` in pytest.

# Issue: {3}.""".format(
#         version, date, will_be, message
#     )
#     warnings.warn(warning, category, stacklevel=stacklevel + 1)


# # Sentinel object for catching pass-through values
# class Unspecified(object):
#     pass


def regularize_path(path):
    """
    Converts pathlib Paths into plain string paths (for all versions of Python).
    """
    is_path = False

    if isinstance(path, getattr(os, "PathLike", ())):
        is_path = True
        path = os.fspath(path)

    elif hasattr(path, "__fspath__"):
        is_path = True
        path = path.__fspath__()

    elif path.__class__.__module__ == "pathlib":
        import pathlib

        if isinstance(path, pathlib.Path):
            is_path = True
            path = str(path)

    return is_path, path


class Behavior(Mapping):
    def __init__(self, defaults, overrides):
        self.defaults = defaults
        if overrides is None:
            self.overrides = {}
        else:
            self.overrides = overrides

    def __getitem__(self, where):
        try:
            return self.overrides[where]
        except KeyError:
            try:
                return self.defaults[where]
            except KeyError:
                return None

    def items(self):
        for n, x in self.overrides.items():
            yield n, x
        for n, x in self.defaults.items():
            if n not in self.overrides:
                yield n, x

    def __iter__(self):
        yield from self.keys()

    def __len__(self):
        return len(set(self.defaults) | set(self.overrides))


def arrayclass(layout, behavior):
    behavior = Behavior(ak._v2.behavior, behavior)
    arr = layout.parameter("__array__")
    if isstr(arr):
        cls = behavior[arr]
        if isinstance(cls, type) and issubclass(cls, ak._v2.highlevel.Array):
            return cls
    rec = layout.parameter("__record__")
    if isstr(rec):
        cls = behavior[".", rec]
        if isinstance(cls, type) and issubclass(cls, ak._v2.highlevel.Array):
            return cls
    deeprec = layout.purelist_parameter("__record__")
    if isstr(deeprec):
        cls = behavior["*", deeprec]
        if isinstance(cls, type) and issubclass(cls, ak._v2.highlevel.Array):
            return cls
    return ak._v2.highlevel.Array


def custom_cast(obj, behavior):
    behavior = Behavior(ak._v2.behavior, behavior)
    for key, fcn in behavior.items():
        if (
            isinstance(key, tuple)
            and len(key) == 2
            and key[0] == "__cast__"
            and isinstance(obj, key[1])
        ):
            return fcn
    return None


def custom_broadcast(layout, behavior):
    behavior = Behavior(ak._v2.behavior, behavior)
    custom = layout.parameter("__array__")
    if not isstr(custom):
        custom = layout.parameter("__record__")
    if not isstr(custom):
        custom = layout.purelist_parameter("__record__")
    if isstr(custom):
        for key, fcn in behavior.items():
            if (
                isinstance(key, tuple)
                and len(key) == 2
                and key[0] == "__broadcast__"
                and key[1] == custom
            ):
                return fcn
    return None


def custom_ufunc(ufunc, layout, behavior):
    import numpy

    behavior = Behavior(ak._v2.behavior, behavior)
    custom = layout.parameter("__array__")
    if not isstr(custom):
        custom = layout.parameter("__record__")
    if not isstr(custom):
        custom = layout.purelist_parameter("__record__")
    if isstr(custom):
        for key, fcn in behavior.items():
            if (
                isinstance(key, tuple)
                and len(key) == 2
                and (key[0] is ufunc or key[0] is numpy.ufunc)
                and key[1] == custom
            ):
                return fcn
    return None


def numba_array_typer(layouttype, behavior):
    behavior = Behavior(ak._v2.behavior, behavior)
    arr = layouttype.parameters.get("__array__")
    if isstr(arr):
        typer = behavior["__numba_typer__", arr]
        if callable(typer):
            return typer
    rec = layouttype.parameters.get("__record__")
    if isstr(rec):
        typer = behavior["__numba_typer__", ".", rec]
        if callable(typer):
            return typer
    deeprec = layouttype.parameters.get("__record__")
    if isstr(deeprec):
        typer = behavior["__numba_typer__", "*", deeprec]
        if callable(typer):
            return typer
    return None


def numba_array_lower(layouttype, behavior):
    behavior = Behavior(ak._v2.behavior, behavior)
    arr = layouttype.parameters.get("__array__")
    if isstr(arr):
        lower = behavior["__numba_lower__", arr]
        if callable(lower):
            return lower
    rec = layouttype.parameters.get("__record__")
    if isstr(rec):
        lower = behavior["__numba_lower__", ".", rec]
        if callable(lower):
            return lower
    deeprec = layouttype.parameters.get("__record__")
    if isstr(deeprec):
        lower = behavior["__numba_lower__", "*", deeprec]
        if callable(lower):
            return lower
    return None


def recordclass(layout, behavior):
    behavior = Behavior(ak._v2.behavior, behavior)
    rec = layout.parameter("__record__")
    if isstr(rec):
        cls = behavior[rec]
        if isinstance(cls, type) and issubclass(cls, ak._v2.highlevel.Record):
            return cls
    return ak._v2.highlevel.Record


def typestrs(behavior):
    behavior = Behavior(ak._v2.behavior, behavior)
    out = {}
    for key, typestr in behavior.items():
        if (
            isinstance(key, tuple)
            and len(key) == 2
            and key[0] == "__typestr__"
            and isstr(key[1])
            and isstr(typestr)
        ):
            out[key[1]] = typestr
    return out


def gettypestr(parameters, typestrs):
    if parameters is not None:
        record = parameters.get("__record__")
        if record is not None:
            typestr = typestrs.get(record)
            if typestr is not None:
                return typestr
        array = parameters.get("__array__")
        if array is not None:
            typestr = typestrs.get(array)
            if typestr is not None:
                return typestr
    return None


def numba_record_typer(layouttype, behavior):
    behavior = Behavior(ak._v2.behavior, behavior)
    rec = layouttype.parameters.get("__record__")
    if isstr(rec):
        typer = behavior["__numba_typer__", rec]
        if callable(typer):
            return typer
    return None


def numba_record_lower(layouttype, behavior):
    behavior = Behavior(ak._v2.behavior, behavior)
    rec = layouttype.parameters.get("__record__")
    if isstr(rec):
        lower = behavior["__numba_lower__", rec]
        if callable(lower):
            return lower
    return None


def overload(behavior, signature):
    if not any(s is None for s in signature):
        behavior = Behavior(ak._v2.behavior, behavior)
        for key, custom in behavior.items():
            if (
                isinstance(key, tuple)
                and len(key) == len(signature)
                and key[0] == signature[0]
                and all(
                    k == s
                    or (
                        isinstance(k, type) and isinstance(s, type) and issubclass(s, k)
                    )
                    for k, s in zip(key[1:], signature[1:])
                )
            ):
                return custom


def numba_attrs(layouttype, behavior):
    behavior = Behavior(ak._v2.behavior, behavior)
    rec = layouttype.parameters.get("__record__")
    if isstr(rec):
        for key, typer in behavior.items():
            if (
                isinstance(key, tuple)
                and len(key) == 3
                and key[0] == "__numba_typer__"
                and key[1] == rec
            ):
                lower = behavior["__numba_lower__", key[1], key[2]]
                yield key[2], typer, lower


def numba_methods(layouttype, behavior):
    behavior = Behavior(ak._v2.behavior, behavior)
    rec = layouttype.parameters.get("__record__")
    if isstr(rec):
        for key, typer in behavior.items():
            if (
                isinstance(key, tuple)
                and len(key) == 4
                and key[0] == "__numba_typer__"
                and key[1] == rec
                and key[3] == ()
            ):
                lower = behavior["__numba_lower__", key[1], key[2], ()]
                yield key[2], typer, lower


def numba_unaryops(unaryop, left, behavior):
    behavior = Behavior(ak._v2.behavior, behavior)
    done = False

    if isinstance(left, ak._v2._connect.numba.layout.ContentType):
        left = left.parameters.get("__record__")
        if not isstr(left):
            done = True

    if not done:
        for key, typer in behavior.items():
            if (
                isinstance(key, tuple)
                and len(key) == 3
                and key[0] == "__numba_typer__"
                and key[1] == unaryop
                and key[2] == left
            ):
                lower = behavior["__numba_lower__", key[1], key[2]]
                yield typer, lower


def numba_binops(binop, left, right, behavior):
    behavior = Behavior(ak._v2.behavior, behavior)
    done = False

    if isinstance(left, ak._v2._connect.numba.layout.ContentType):
        left = left.parameters.get("__record__")
        if not isstr(left):
            done = True

    if isinstance(right, ak._v2._connect.numba.layout.ContentType):
        right = right.parameters.get("__record__")
        if not isstr(right):
            done = True

    if not done:
        for key, typer in behavior.items():
            if (
                isinstance(key, tuple)
                and len(key) == 4
                and key[0] == "__numba_typer__"
                and key[1] == left
                and key[2] == binop
                and key[3] == right
            ):
                lower = behavior["__numba_lower__", key[1], key[2], key[3]]
                yield typer, lower


def behavior_of(*arrays, **kwargs):
    behavior = kwargs.get("behavior")
    if behavior is not None:
        # An explicit 'behavior' always wins.
        return behavior

    copied = False
    highs = (
        ak._v2.highlevel.Array,
        ak._v2.highlevel.Record,
        #        ak._v2.highlevel.ArrayBuilder,
    )
    for x in arrays[::-1]:
        if isinstance(x, highs) and x.behavior is not None:
            if behavior is None:
                behavior = x.behavior
            elif behavior is x.behavior:
                pass
            elif not copied:
                behavior = dict(behavior)
                behavior.update(x.behavior)
                copied = True
            else:
                behavior.update(x.behavior)
    return behavior


# maybe_wrap and maybe_wrap_like go here
def wrap(content, behavior=None, highlevel=True, like=None):
    assert content is None or isinstance(
        content, (ak._v2.contents.Content, ak._v2.record.Record)
    )
    assert behavior is None or isinstance(behavior, Mapping)
    assert isinstance(highlevel, bool)
    if highlevel:
        if like is not None and behavior is None:
            behavior = behavior_of(like)

        if isinstance(content, ak._v2.contents.Content):
            return ak._v2.highlevel.Array(content, behavior=behavior)
        elif isinstance(content, ak._v2.record.Record):
            return ak._v2.highlevel.Record(content, behavior=behavior)

    return content


def extra(args, kwargs, defaults):
    out = []
    for i, (name, default) in enumerate(defaults):
        if i < len(args):
            out.append(args[i])
        elif name in kwargs:
            out.append(kwargs[name])
        else:
            out.append(default)
    return out


# def key2index(keys, key):
#     if keys is None:
#         attempt = None
#     else:
#         try:
#             attempt = keys.index(key)
#         except ValueError:
#             attempt = None

#     if attempt is None:
#         m = key2index._pattern.match(key)
#         if m is not None:
#             attempt = m.group(0)

#     if attempt is None:
#         raise ValueError(
#             "key {0} not found in record".format(repr(key))
#         )
#     else:
#         return attempt


# key2index._pattern = re.compile(r"^[1-9][0-9]*$")


# def make_union(tags, index, contents, identities, parameters):
#     if isinstance(index, ak._v2.contents.Index32):
#         return ak._v2.contents.UnionArray8_32(tags, index, contents, identities, parameters)
#     elif isinstance(index, ak._v2.contents.IndexU32):
#         return ak._v2.contents.UnionArray8_U32(tags, index, contents, identities, parameters)
#     elif isinstance(index, ak._v2.index.Index64):
#         return ak._v2.contents.UnionArray8_64(tags, index, contents, identities, parameters)
#     else:
#         raise AssertionError(index)


# def union_to_record(unionarray, anonymous):
#     nplike = ak.nplike.of(unionarray)

#     contents = []
#     for layout in unionarray.contents:
#         if isinstance(layout, virtualtypes):
#             contents.append(layout.array)
#         elif isinstance(layout, indexedtypes):
#             contents.append(layout.project())
#         elif isinstance(layout, uniontypes):
#             contents.append(union_to_record(layout, anonymous))
#         elif isinstance(layout, optiontypes):
#             contents.append(
#                 ak._v2.operations.structure.fill_none(
#                     layout, np.nan, axis=0, highlevel=False
#                 )
#             )
#         else:
#             contents.append(layout)

#     if not any(isinstance(x, ak._v2.contents.RecordArray) for x in contents):
#         return make_union(
#             unionarray.tags,
#             unionarray.index,
#             contents,
#             unionarray.identities,
#             unionarray.parameters,
#         )

#     else:
#         seen = set()
#         all_names = []
#         for layout in contents:
#             if isinstance(layout, ak._v2.contents.RecordArray):
#                 for key in layout.keys():
#                     if key not in seen:
#                         seen.add(key)
#                         all_names.append(key)
#             else:
#                 if anonymous not in seen:
#                     seen.add(anonymous)
#                     all_names.append(anonymous)

#         missingarray = ak._v2.contents.IndexedOptionArray64(
#             ak._v2.index.Index64(nplike.full(len(unionarray), -1, dtype=np.int64)),
#             ak._v2.contents.EmptyArray(),
#         )

#         all_fields = []
#         for name in all_names:
#             union_contents = []
#             for layout in contents:
#                 if isinstance(layout, ak._v2.contents.RecordArray):
#                     for key in layout.keys():
#                         if name == key:
#                             union_contents.append(layout.field(key))
#                             break
#                     else:
#                         union_contents.append(missingarray)
#                 else:
#                     if name == anonymous:
#                         union_contents.append(layout)
#                     else:
#                         union_contents.append(missingarray)

#             all_fields.append(
#                 make_union(
#                     unionarray.tags,
#                     unionarray.index,
#                     union_contents,
#                     unionarray.identities,
#                     unionarray.parameters,
#                 ).simplify()
#             )

#         return ak._v2.contents.RecordArray(all_fields, all_names, len(unionarray))


def direct_Content_subclass(node):
    if node is None:
        return None
    else:
        mro = type(node).mro()
        return mro[mro.index(ak._v2.contents.Content) - 1]


def direct_Content_subclass_name(node):
    out = direct_Content_subclass(node)
    if out is None:
        return None
    else:
        return out.__name__


def merge_parameters(one, two, merge_equal=False):

    if one is None and two is None:
        return None

    elif one is None:
        return two

    elif two is None:
        return one

    elif merge_equal:
        out = {}
        for k, v in two.items():
            if k in one.keys():
                if v == one[k]:
                    out[k] = v
        return out

    else:
        out = dict(one)
        for k, v in two.items():
            if v is not None:
                out[k] = v
        return out



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


def is_none(array, axis=0, highlevel=True, behavior=None):
    """
    Args:
        array: Data to check for missing values (None).
        axis (int): The dimension at which this operation is applied. The
            outermost dimension is `0`, followed by `1`, etc., and negative
            values count backward from the innermost: `-1` is the innermost
            dimension, `-2` is the next level up, etc.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.layout.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.

    Returns an array whose value is True where an element of `array` is None;
    False otherwise (at a given `axis` depth).
    """

    # Determine the (potentially nested) bytemask
    def getfunction_inner(layout, depth, **kwargs):
        if not isinstance(layout, ak._v2.contents.Content):
            return

        nplike = ak.nplike.of(layout)

        if layout.is_OptionType:
            layout = layout.toIndexedOptionArray64()

            # Convert the option type into a union, using the mask
            # as a tag.
            tag = nplike.asarray(layout.bytemask())
            index = nplike.where(tag, 0, nplike.asarray(layout.index))

            return ak._v2.contents.UnionArray(
                ak._v2.index.Index8(tag),
                ak._v2.index.Index64(index),
                [
                    layout.content.recursively_apply(getfunction_inner),
                    ak._v2.contents.NumpyArray(nplike.array([True], dtype=np.bool_)),
                ],
            ).simplify_uniontype()

        elif (
            layout.is_UnknownType
            or layout.is_ListType
            or layout.is_RecordType
            or layout.is_NumpyType
        ):
            return ak._v2.contents.NumpyArray(nplike.zeros(len(layout), dtype=np.bool_))

    # Locate the axis
    def getfunction_outer(layout, depth, depth_context, **kwargs):
        depth_context["posaxis"] = layout.axis_wrap_if_negative(
            depth_context["posaxis"]
        )
        if depth_context["posaxis"] == depth - 1:
            return layout.recursively_apply(getfunction_inner)

    layout = ak._v2.operations.convert.to_layout(array)
    behavior = ak._v2._util.behavior_of(array, behavior=behavior)
    depth_context = {"posaxis": axis}
    out = layout.recursively_apply(getfunction_outer, depth_context=depth_context)
    return ak._v2._util.wrap(out, behavior, highlevel)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


def to_regular(array, axis=1, highlevel=True, behavior=None):
    """
    Args:
        array: Array-like data (anything #ak.to_layout recognizes).
        axis (int or None): The dimension at which this operation is applied.
            The outermost dimension is `0`, followed by `1`, etc., and negative
            values count backward from the innermost: `-1` is the innermost
            dimension, `-2` is the next level up, etc. If None, convert all
            variable dimensions into regular ones or raise a ValueError if that
            is not possible.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.layout.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.

    Converts a variable-length axis into a regular one, if possible.

        >>> irregular = ak.from_iter(np.arange(2*3*5).reshape(2, 3, 5))
        >>> print(irregular.type)
        2 * var * var * int64
        >>> print(ak.to_regular(irregular).type)
        2 * 3 * var * int64
        >>> print(ak.to_regular(irregular, axis=2).type)
        2 * var * 5 * int64
        >>> print(ak.to_regular(irregular, axis=-1).type)
        2 * var * 5 * int64

    But truly irregular data cannot be converted.

        >>> ak.to_regular(ak.Array([[1, 2, 3], [], [4, 5]]))
        ValueError: in ListOffsetArray64, cannot convert to RegularArray because
        subarray lengths are not regular

    See also #ak.from_regular.
    """
    layout = ak._v2.operations.convert.to_layout(array)
    posaxis = layout.axis_wrap_if_negative(axis)

    if axis is None:

        def action(layout, continuation, **kwargs):
            if layout.is_ListType:
                return continuation().toRegularArray()

        out = layout.recursively_apply(action)

    elif posaxis == 0:
        out = layout  # the top-level can only be regular (ArrayType)

    else:

        def action(layout, depth, depth_context, **kwargs):
            posaxis = layout.axis_wrap_if_negative(depth_context["posaxis"])
            if posaxis == depth and layout.is_ListType:
                return layout.toRegularArray()
            elif posaxis == 0:
                raise np.AxisError(
                    f"axis={axis} exceeds the depth of this array ({depth})"
                )

            depth_context["posaxis"] = posaxis

        depth_context = {"posaxis": posaxis}
        out = layout.recursively_apply(action, depth_context)

    return ak._v2._util.wrap(out, behavior, highlevel)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE


import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401

to_list = ak._v2.operations.convert.to_list


def test_one_level():
    base = ak._v2.operations.structure.zip(
        {"a": ak._v2.operations.structure.zip({"x": [1, 2, 3]}), "b": [1, 2, 3]},
        depth_limit=1,
    )
    what = ak._v2.Array([1.1, 2.2, 3.3], check_valid=True)
    assert to_list(
        ak._v2.operations.structure.with_field(base, what, where=["a", "y"])
    ) == [
        {"b": 1, "a": {"x": 1, "y": 1.1}},
        {"b": 2, "a": {"x": 2, "y": 2.2}},
        {"b": 3, "a": {"x": 3, "y": 3.3}},
    ]
    assert ak._v2.operations.describe.fields(base) == ["a", "b"]

    base["a", "y"] = what
    assert to_list(base) == [
        {"b": 1, "a": {"x": 1, "y": 1.1}},
        {"b": 2, "a": {"x": 2, "y": 2.2}},
        {"b": 3, "a": {"x": 3, "y": 3.3}},
    ]


def test_two_level():
    base = ak._v2.operations.structure.zip(
        {
            "A": ak._v2.operations.structure.zip(
                {"a": ak._v2.operations.structure.zip({"x": [1, 2, 3]}), "b": [1, 2, 3]}
            ),
            "B": [1, 2, 3],
        },
        depth_limit=1,
    )
    what = ak._v2.Array([1.1, 2.2, 3.3], check_valid=True)
    assert to_list(
        ak._v2.operations.structure.with_field(base, what, where=["A", "a", "y"])
    ) == [
        {"B": 1, "A": {"b": 1, "a": {"x": 1, "y": 1.1}}},
        {"B": 2, "A": {"b": 2, "a": {"x": 2, "y": 2.2}}},
        {"B": 3, "A": {"b": 3, "a": {"x": 3, "y": 3.3}}},
    ]
    assert ak._v2.operations.describe.fields(base) == ["A", "B"]

    base["A", "a", "y"] = what
    assert to_list(base) == [
        {"B": 1, "A": {"b": 1, "a": {"x": 1, "y": 1.1}}},
        {"B": 2, "A": {"b": 2, "a": {"x": 2, "y": 2.2}}},
        {"B": 3, "A": {"b": 3, "a": {"x": 3, "y": 3.3}}},
    ]


def test_replace_the_only_field():
    base = ak._v2.operations.structure.zip(
        {"a": ak._v2.operations.structure.zip({"x": [1, 2, 3]})}, depth_limit=1
    )
    what = ak._v2.Array([1.1, 2.2, 3.3], check_valid=True)
    assert to_list(
        ak._v2.operations.structure.with_field(base, what, where=["a", "y"])
    ) == [
        {"a": {"x": 1, "y": 1.1}},
        {"a": {"x": 2, "y": 2.2}},
        {"a": {"x": 3, "y": 3.3}},
    ]
    assert ak._v2.operations.describe.fields(base) == ["a"]

    base["a", "y"] = what
    assert to_list(base) == [
        {"a": {"x": 1, "y": 1.1}},
        {"a": {"x": 2, "y": 2.2}},
        {"a": {"x": 3, "y": 3.3}},
    ]


def test_check_no_field():
    base = ak._v2.operations.structure.zip(
        {"a": ak._v2.operations.structure.zip({"x": [1, 2, 3]})}, depth_limit=1
    )
    what = ak._v2.Array([1.1, 2.2, 3.3], check_valid=True)

    assert to_list(ak._v2.operations.structure.with_field(base, what)) == [
        {"a": {"x": 1}, "1": 1.1},
        {"a": {"x": 2}, "1": 2.2},
        {"a": {"x": 3}, "1": 3.3},
    ]

    with pytest.raises(ValueError):
        ak._v2.operations.structure.with_field(what, what)

    content1 = ak._v2.contents.NumpyArray(np.array([1, 2, 3]))
    recordarray = ak._v2.contents.RecordArray([content1], None)
    what = ak._v2.Array([1.1, 2.2, 3.3], check_valid=True)

    assert to_list(ak._v2.operations.structure.with_field(recordarray, what)) == [
        (1, 1.1),
        (2, 2.2),
        (3, 3.3),
    ]
    assert to_list(ak._v2.operations.structure.with_field(recordarray, what, "a")) == [
        {"0": 1, "a": 1.1},
        {"0": 2, "a": 2.2},
        {"0": 3, "a": 3.3},
    ]
    assert ak._v2.operations.describe.fields(recordarray) == ["0"]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE


import pytest  # noqa: F401
import numpy as np  # noqa: F401
import cupy as cp  # noqa: F401
import awkward._v2 as ak  # noqa: F401


def test_cupy_interop():
    c = cp.arange(10)
    n = np.arange(10)
    cupy_index_arr = ak.index.Index64(c)
    np_index_arr = ak.index.Index64(n)

    # GPU->CPU
    assert ak.to_list(np.asarray(cupy_index_arr.to_backend("cpu"))) == ak.to_list(
        np.asarray(np_index_arr)
    )
    # CPU->CPU
    assert ak.to_list(np.asarray(np_index_arr.to_backend("cpu"))) == ak.to_list(
        np.asarray(np_index_arr)
    )
    # CPU->GPU->CPU
    assert ak.to_list(np.asarray(np_index_arr)) == ak.to_list(
        np.asarray(np_index_arr.to_backend("cuda").to_backend("cpu"))
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from collections.abc import Iterable
import copy

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


def with_field(base, what, where=None, highlevel=True, behavior=None):

    """
    Args:
        base: Data containing records or tuples.
        what: Data to add as a new field.
        where (None or str or non-empy iterable of str): If None, the new field
            has no name (can be accessed as an integer slot number in a
            string); If str, the name of the new field. If iterable, it is
            interpreted as a path where to add the field in a nested record.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.layout.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.

    Returns an #ak.Array or #ak.Record (or low-level equivalent, if
    `highlevel=False`) with a new field attached. This function does not
    change the array in-place.

    See #ak.Array.__setitem__ and #ak.Record.__setitem__ for a variant that
    changes the high-level object in-place. (These methods internally use
    #ak.with_field, so performance is not a factor in choosing one over the
    other.)
    """

    if not (
        where is None
        or isinstance(where, str)
        or (isinstance(where, Iterable) and all(isinstance(x, str) for x in where))
    ):
        raise TypeError(
            "New fields may only be assigned by field name(s) "
            "or as a new integer slot by passing None for 'where'"
        )
    if (
        not isinstance(where, str)
        and isinstance(where, Iterable)
        and all(isinstance(x, str) for x in where)
        and len(where) > 1
    ):
        return with_field(
            base,
            with_field(
                base[where[0]],
                what,
                where=where[1:],
                highlevel=highlevel,
                behavior=behavior,
            ),
            where=where[0],
            highlevel=highlevel,
            behavior=behavior,
        )
    else:

        if not (isinstance(where, str) or where is None):
            where = where[0]

        behavior = ak._v2._util.behavior_of(base, what, behavior=behavior)
        base = ak._v2.operations.convert.to_layout(
            base, allow_record=True, allow_other=False
        )

        if len(base.fields) == 0:
            raise ValueError("no tuples or records in array; cannot add a new field")

        what = ak._v2.operations.convert.to_layout(
            what, allow_record=True, allow_other=True
        )

        keys = copy.copy(base.fields)
        if where in base.fields:
            keys.remove(where)

        if len(keys) == 0:
            # the only key was removed, so just create new Record
            out = (
                ak._v2.contents.RecordArray(
                    [what], [where], parameters=base.parameters
                ),
            )

        else:

            def action(inputs, **kwargs):
                nplike = ak.nplike.of(*inputs)
                base, what = inputs
                if isinstance(base, ak._v2.contents.RecordArray):
                    if what is None:
                        what = ak._v2.contents.IndexedOptionArray(
                            ak._v2.index.Index64(nplike.full(len(base), -1, np.int64)),
                            ak._v2.contents.EmptyArray(),
                        )
                    elif not isinstance(what, ak._v2.contents.Content):
                        what = ak._v2.contents.NumpyArray(
                            nplike.repeat(what, len(base))
                        )
                    if base.is_tuple and where is None:
                        fields = None
                    elif base.is_tuple:
                        fields = keys + [where]
                    elif where is None:
                        fields = keys + [str(len(keys))]
                    else:
                        fields = keys + [where]
                    out = ak._v2.contents.RecordArray(
                        [base[k] for k in keys] + [what],
                        fields,
                        parameters=base.parameters,
                    )
                    return (out,)
                else:
                    return None

            out = ak._v2._broadcasting.broadcast_and_apply(
                [base, what],
                action,
                behavior,
                right_broadcast=False,
            )

        assert isinstance(out, tuple) and len(out) == 1

        return ak._v2._util.wrap(out[0], behavior, highlevel)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from collections.abc import Iterable

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()
numpy = ak.nplike.Numpy.instance()


def to_layout(
    array,
    allow_record=True,
    allow_other=False,
    numpytype=(np.number, np.bool_, np.str_, np.bytes_, np.datetime64, np.timedelta64),
):
    """
    Args:
        array: Data to convert into a low-level #ak.layout.Content layout
            or maybe #ak.layout.Record, its record equivalent, or other types.
        allow_record (bool): If True, allow #ak.layout.Record as an output;
            otherwise, if the output would be a scalar record, raise an error.
        allow_other (bool): If True, allow non-Awkward outputs; otherwise,
            if the output would be another type, raise an error.
        numpytype (tuple of NumPy types): Dtypes to allow from NumPy arrays.

    Converts `array` (many types supported, including all Awkward Arrays and
    Records) into a #ak.layout.Content and maybe #ak.layout.Record or
    other types.

    This function is usually used to sanitize inputs for other functions; it
    would rarely be used in a data analysis because #ak.layout.Content and
    #ak.layout.Record are lower-level than #ak.Array.
    """
    if isinstance(array, ak._v2.contents.Content):
        return array

    elif allow_record and isinstance(array, ak._v2.record.Record):
        return array

    elif isinstance(array, ak._v2.highlevel.Array):
        return array.layout

    elif allow_record and isinstance(array, ak._v2.highlevel.Record):
        return array.layout

    # elif isinstance(array, ak._v2.highlevel.ArrayBuilder):
    #     return array.snapshot().layout

    # elif isinstance(array, ak.layout.ArrayBuilder):
    #     return array.snapshot()

    elif isinstance(array, (np.ndarray, numpy.ma.MaskedArray)):
        if not issubclass(array.dtype.type, numpytype):
            raise ValueError(f"dtype {array.dtype!r} not allowed")
        return to_layout(
            ak._v2.operations.convert.from_numpy(
                array, regulararray=True, recordarray=True, highlevel=False
            ),
            allow_record=allow_record,
            allow_other=allow_other,
        )

    elif (
        type(array).__module__.startswith("cupy.") and type(array).__name__ == "ndarray"
    ):
        if not issubclass(array.dtype.type, numpytype):
            raise ValueError(f"dtype {array.dtype!r} not allowed")
        return to_layout(
            ak._v2.operations.convert.from_cupy(
                array, regulararray=True, highlevel=False
            ),
            allow_record=allow_record,
            allow_other=allow_other,
        )

    elif isinstance(array, (str, bytes)):
        return to_layout(
            ak._v2.operations.convert.from_iter([array], highlevel=False),
            allow_record=allow_record,
            allow_other=allow_other,
        )

    elif isinstance(array, Iterable):
        return to_layout(
            ak._v2.operations.convert.from_iter(array, highlevel=False),
            allow_record=allow_record,
            allow_other=allow_other,
        )

    elif not allow_other:
        raise TypeError(f"{array} cannot be converted into an Awkward Array")

    else:
        return array



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE


import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    assert ak.to_list(
        ak.layout.NumpyArray(np.array([[1, 2], [3, 4]], np.int64)).numbers_to_type(
            "int16"
        )
    ) == [[1, 2], [3, 4]]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


def from_arrow(array, conservative_optiontype=False, highlevel=True, behavior=None):
    """
    Args:
        array (`pyarrow.Array`, `pyarrow.ChunkedArray`, `pyarrow.RecordBatch`,
            or `pyarrow.Table`): Apache Arrow array to convert into an
            Awkward Array.
        conservative_optiontype (bool): If enabled and the optionalness of a type
            can't be determined (i.e. not within an `pyarrow.field` and the Arrow array
            has no Awkward metadata), assume that it is option-type with a blank
            BitMaskedArray.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.layout.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.

    Converts an Apache Arrow array into an Awkward Array.

    This function always preserves the values of a dataset; i.e. the Python objects
    returned by #ak.to_list are identical to the Python objects returned by Arrow's
    `to_pylist` method. If #ak.to_arrow was invoked with `extensionarray=True`, this
    function also preserves the data type (high-level #ak.types.Type, though not the
    low-level #ak.forms.Form), even through Parquet, making Parquet a good way to save
    Awkward Arrays for later use.

    See also #ak.to_arrow, #ak.to_arrow_table, #ak.from_parquet, #ak.from_arrow_schema.
    """
    import awkward._v2._connect.pyarrow

    out = awkward._v2._connect.pyarrow.handle_arrow(
        array, conservative_optiontype=conservative_optiontype, pass_empty_field=True
    )
    return ak._v2._util.wrap(out, behavior, highlevel)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import json
import copy
import math

import awkward as ak
from awkward._v2.index import Index
from awkward._v2._slicing import NestedIndexError
from awkward._v2.contents.content import Content
from awkward._v2.contents.bytemaskedarray import ByteMaskedArray
from awkward._v2.forms.bitmaskedform import BitMaskedForm
from awkward._v2.forms.form import _parameters_equal

np = ak.nplike.NumpyMetadata.instance()
numpy = ak.nplike.Numpy.instance()


class BitMaskedArray(Content):
    is_OptionType = True

    def __init__(
        self,
        mask,
        content,
        valid_when,
        length,
        lsb_order,
        identifier=None,
        parameters=None,
        nplike=None,
    ):
        if not (isinstance(mask, Index) and mask.dtype == np.dtype(np.uint8)):
            raise TypeError(
                "{} 'mask' must be an Index with dtype=uint8, not {}".format(
                    type(self).__name__, repr(mask)
                )
            )
        if not isinstance(content, Content):
            raise TypeError(
                "{} 'content' must be a Content subtype, not {}".format(
                    type(self).__name__, repr(content)
                )
            )
        if not isinstance(valid_when, bool):
            raise TypeError(
                "{} 'valid_when' must be boolean, not {}".format(
                    type(self).__name__, repr(valid_when)
                )
            )
        if not isinstance(length, ak._v2._typetracer.UnknownLengthType):
            if not (ak._util.isint(length) and length >= 0):
                raise TypeError(
                    "{} 'length' must be a non-negative integer, not {}".format(
                        type(self).__name__, length
                    )
                )
        if not isinstance(lsb_order, bool):
            raise TypeError(
                "{} 'lsb_order' must be boolean, not {}".format(
                    type(self).__name__, repr(lsb_order)
                )
            )
        if length > mask.length * 8:
            raise ValueError(
                "{} 'length' ({}) must be <= len(mask) * 8 ({})".format(
                    type(self).__name__, length, mask.length * 8
                )
            )
        if length > content.length:
            raise ValueError(
                "{} 'length' ({}) must be <= len(content) ({})".format(
                    type(self).__name__, length, content.length
                )
            )
        if nplike is None:
            nplike = content.nplike
        if nplike is None:
            nplike = mask.nplike

        self._mask = mask
        self._content = content
        self._valid_when = valid_when
        self._length = length
        self._lsb_order = lsb_order
        self._init(identifier, parameters, nplike)

    @property
    def mask(self):
        return self._mask

    @property
    def content(self):
        return self._content

    @property
    def valid_when(self):
        return self._valid_when

    @property
    def lsb_order(self):
        return self._lsb_order

    Form = BitMaskedForm

    def _form_with_key(self, getkey):
        form_key = getkey(self)
        return self.Form(
            self._mask.form,
            self._content._form_with_key(getkey),
            self._valid_when,
            self._lsb_order,
            has_identifier=self._identifier is not None,
            parameters=self._parameters,
            form_key=form_key,
        )

    def _to_buffers(self, form, getkey, container, nplike):
        assert isinstance(form, self.Form)
        key = getkey(self, form, "mask")
        container[key] = ak._v2._util.little_endian(self._mask.raw(nplike))
        self._content._to_buffers(form.content, getkey, container, nplike)

    @property
    def typetracer(self):
        tt = ak._v2._typetracer.TypeTracer.instance()
        return BitMaskedArray(
            ak._v2.index.Index(self._mask.raw(tt)),
            self._content.typetracer,
            self._valid_when,
            self._length,
            self._lsb_order,
            self._typetracer_identifier(),
            self._parameters,
            tt,
        )

    @property
    def length(self):
        return self._length

    def _forget_length(self):
        return BitMaskedArray(
            self._mask,
            self._content.typetracer,
            self._valid_when,
            ak._v2._typetracer.UnknownLength,
            self._lsb_order,
            self._identifier,
            self._parameters,
            self._nplike,
        )

    def __repr__(self):
        return self._repr("", "", "")

    def _repr(self, indent, pre, post):
        out = [indent, pre, "<BitMaskedArray valid_when="]
        out.append(repr(json.dumps(self._valid_when)))
        out.append(" lsb_order=")
        out.append(repr(json.dumps(self._lsb_order)))
        out.append(" len=")
        out.append(repr(str(self.length)))
        out.append(">")
        out.extend(self._repr_extra(indent + "    "))
        out.append("\n")
        out.append(self._mask._repr(indent + "    ", "<mask>", "</mask>\n"))
        out.append(self._content._repr(indent + "    ", "<content>", "</content>\n"))
        out.append(indent + "</BitMaskedArray>")
        out.append(post)
        return "".join(out)

    def merge_parameters(self, parameters):
        return BitMaskedArray(
            self._mask,
            self._content,
            self._valid_when,
            self._length,
            self._lsb_order,
            self._identifier,
            ak._v2._util.merge_parameters(self._parameters, parameters),
            self._nplike,
        )

    def toByteMaskedArray(self):
        bytemask = ak._v2.index.Index8.empty(self._mask.length * 8, self._nplike)
        assert bytemask.nplike is self._nplike and self._mask.nplike is self._nplike
        self._handle_error(
            self._nplike[
                "awkward_BitMaskedArray_to_ByteMaskedArray",
                bytemask.dtype.type,
                self._mask.dtype.type,
            ](
                bytemask.data,
                self._mask.data,
                self._mask.length,
                False,  # this differs from the kernel call in 'bytemask'
                self._lsb_order,
            )
        )
        return ByteMaskedArray(
            bytemask[: self._length],
            self._content,
            self._valid_when,
            self._identifier,
            self._parameters,
            self._nplike,
        )

    def toIndexedOptionArray64(self):
        index = ak._v2.index.Index64.empty(self._mask.length * 8, self._nplike)
        assert index.nplike is self._nplike and self._mask.nplike is self._nplike
        self._handle_error(
            self._nplike[
                "awkward_BitMaskedArray_to_IndexedOptionArray",
                index.dtype.type,
                self._mask.dtype.type,
            ](
                index.raw(self._nplike),
                self._mask.data,
                self._mask.length,
                self._valid_when,
                self._lsb_order,
            ),
        )
        return ak._v2.contents.indexedoptionarray.IndexedOptionArray(
            index[0 : self._length],
            self._content,
            self._identifier,
            self._parameters,
            self._nplike,
        )

    def mask_as_bool(self, valid_when=None, nplike=None):
        if valid_when is None:
            valid_when = self._valid_when
        if nplike is None:
            nplike = self._nplike

        bytemask = ak._v2.index.Index8.empty(self._mask.length * 8, nplike)
        assert bytemask.nplike is self._nplike and self._mask.nplike is self._nplike
        self._handle_error(
            nplike[
                "awkward_BitMaskedArray_to_ByteMaskedArray",
                bytemask.dtype.type,
                self._mask.dtype.type,
            ](
                bytemask.data,
                self._mask.data,
                self._mask.length,
                0 if valid_when == self._valid_when else 1,
                self._lsb_order,
            )
        )
        return bytemask.data[: self._length].view(np.bool_)

    def _getitem_nothing(self):
        return self._content._getitem_range(slice(0, 0))

    def _getitem_at(self, where):
        if not self._nplike.known_data:
            return ak._v2._typetracer.MaybeNone(self._content._getitem_at(where))

        if where < 0:
            where += self.length
        if not (0 <= where < self.length) and self._nplike.known_shape:
            raise NestedIndexError(self, where)
        if self._lsb_order:
            bit = bool(self._mask[where // 8] & (1 << (where % 8)))
        else:
            bit = bool(self._mask[where // 8] & (128 >> (where % 8)))
        if bit == self._valid_when:
            return self._content._getitem_at(where)
        else:
            return None

    def _getitem_range(self, where):
        return self.toByteMaskedArray()._getitem_range(where)

    def _getitem_field(self, where, only_fields=()):
        return BitMaskedArray(
            self._mask,
            self._content._getitem_field(where, only_fields),
            self._valid_when,
            self._length,
            self._lsb_order,
            self._field_identifier(where),
            None,
            self._nplike,
        ).simplify_optiontype()

    def _getitem_fields(self, where, only_fields=()):
        return BitMaskedArray(
            self._mask,
            self._content._getitem_fields(where, only_fields),
            self._valid_when,
            self._length,
            self._lsb_order,
            self._fields_identifier(where),
            None,
            self._nplike,
        ).simplify_optiontype()

    def _carry(self, carry, allow_lazy, exception):
        assert isinstance(carry, ak._v2.index.Index)
        return self.toByteMaskedArray()._carry(carry, allow_lazy, exception)

    def _getitem_next_jagged(self, slicestarts, slicestops, slicecontent, tail):
        return self.toByteMaskedArray()._getitem_next_jagged(
            slicestarts, slicestops, slicecontent, tail
        )

    def _getitem_next(self, head, tail, advanced):
        if head == ():
            return self

        elif isinstance(head, (int, slice, ak._v2.index.Index64)):
            return self.toByteMaskedArray()._getitem_next(head, tail, advanced)

        elif ak._util.isstr(head):
            return self._getitem_next_field(head, tail, advanced)

        elif isinstance(head, list):
            return self._getitem_next_fields(head, tail, advanced)

        elif head is np.newaxis:
            return self._getitem_next_newaxis(tail, advanced)

        elif head is Ellipsis:
            return self._getitem_next_ellipsis(tail, advanced)

        elif isinstance(head, ak._v2.contents.ListOffsetArray):
            return self.toByteMaskedArray()._getitem_next(head, tail, advanced)

        elif isinstance(head, ak._v2.contents.IndexedOptionArray):
            return self._getitem_next_missing(head, tail, advanced)

        else:
            raise AssertionError(repr(head))

    def project(self, mask=None):
        return self.toByteMaskedArray().project(mask)

    def simplify_optiontype(self):
        if isinstance(
            self._content,
            (
                ak._v2.contents.indexedarray.IndexedArray,
                ak._v2.contents.indexedoptionarray.IndexedOptionArray,
                ak._v2.contents.bytemaskedarray.ByteMaskedArray,
                ak._v2.contents.bitmaskedarray.BitMaskedArray,
                ak._v2.contents.unmaskedarray.UnmaskedArray,
            ),
        ):
            return self.toIndexedOptionArray64().simplify_optiontype()
        else:
            return self

    def num(self, axis, depth=0):
        return self.toByteMaskedArray().num(axis, depth)

    def _offsets_and_flattened(self, axis, depth):
        return self.toByteMaskedArray._offsets_and_flattened(axis, depth)

    def mergeable(self, other, mergebool):
        if not _parameters_equal(self._parameters, other._parameters):
            return False

        if isinstance(
            other,
            (
                ak._v2.contents.emptyarray.EmptyArray,
                ak._v2.contents.unionarray.UnionArray,
            ),
        ):
            return True

        if isinstance(
            other,
            (
                ak._v2.contents.indexedarray.IndexedArray,
                ak._v2.contents.indexedoptionarray.IndexedOptionArray,
                ak._v2.contents.bytemaskedarray.ByteMaskedArray,
                ak._v2.contents.bitmaskedarray.BitMaskedArray,
                ak._v2.contents.unmaskedarray.UnmaskedArray,
            ),
        ):
            return self._content.mergeable(other.content, mergebool)

        else:
            return self._content.mergeable(other, mergebool)

    def _reverse_merge(self, other):
        return self.toIndexedOptionArray64()._reverse_merge(other)

    def mergemany(self, others):
        if len(others) == 0:
            return self
        return self.toIndexedOptionArray64().mergemany(others)

    def fillna(self, value):
        return self.toIndexedOptionArray64().fillna(value)

    def _localindex(self, axis, depth):
        return self.toByteMaskedArray()._localindex(axis, depth)

    def numbers_to_type(self, name):
        return self.toByteMaskedArray().numbers_to_type(name)

    def _is_unique(self, negaxis, starts, parents, outlength):
        if self._mask.length == 0:
            return True
        return self.toIndexedOptionArray64()._is_unique(
            negaxis, starts, parents, outlength
        )

    def _unique(self, negaxis, starts, parents, outlength):
        if self._mask.length == 0:
            return self
        out = self.toIndexedOptionArray64()._unique(negaxis, starts, parents, outlength)
        if negaxis is None:
            return out
        else:
            return out._content

    def _argsort_next(
        self,
        negaxis,
        starts,
        shifts,
        parents,
        outlength,
        ascending,
        stable,
        kind,
        order,
    ):
        if self._mask.length == 0:
            return ak._v2.contents.NumpyArray(
                self._nplike.empty(0, np.int64), None, None, self._nplike
            )

        return self.toIndexedOptionArray64()._argsort_next(
            negaxis,
            starts,
            shifts,
            parents,
            outlength,
            ascending,
            stable,
            kind,
            order,
        )

    def _sort_next(
        self, negaxis, starts, parents, outlength, ascending, stable, kind, order
    ):
        return self.toIndexedOptionArray64()._sort_next(
            negaxis,
            starts,
            parents,
            outlength,
            ascending,
            stable,
            kind,
            order,
        )

    def _combinations(self, n, replacement, recordlookup, parameters, axis, depth):
        return self.toByteMaskedArray()._combinations(
            n, replacement, recordlookup, parameters, axis, depth
        )

    def _reduce_next(
        self,
        reducer,
        negaxis,
        starts,
        shifts,
        parents,
        outlength,
        mask,
        keepdims,
    ):
        return self.toByteMaskedArray()._reduce_next(
            reducer,
            negaxis,
            starts,
            shifts,
            parents,
            outlength,
            mask,
            keepdims,
        )

    def _validityerror(self, path):
        if self.mask.length * 8 < self.length:
            return f'at {path} ("{type(self)}"): len(mask) * 8 < length'
        elif self._content.length < self.length:
            return f'at {path} ("{type(self)}"): len(content) < length'
        elif isinstance(
            self._content,
            (
                ak._v2.contents.bitmaskedarray.BitMaskedArray,
                ak._v2.contents.bytemaskedarray.ByteMaskedArray,
                ak._v2.contents.indexedarray.IndexedArray,
                ak._v2.contents.indexedoptionarray.IndexedOptionArray,
                ak._v2.contents.unmaskedarray.UnmaskedArray,
            ),
        ):
            return "{0} contains \"{1}\", the operation that made it might have forgotten to call 'simplify_optiontype()'"
        else:
            return self._content.validityerror(path + ".content")

    def _nbytes_part(self):
        result = self.mask._nbytes_part() + self.content._nbytes_part()
        if self.identifier is not None:
            result = result + self.identifier._nbytes_part()
        return result

    def _rpad(self, target, axis, depth, clip):
        return self.toByteMaskedArray()._rpad(target, axis, depth, clip)

    def _to_arrow(self, pyarrow, mask_node, validbytes, length, options):
        return self.toByteMaskedArray()._to_arrow(
            pyarrow, mask_node, validbytes, length, options
        )

    def _to_numpy(self, allow_missing):
        return self.toByteMaskedArray()._to_numpy(allow_missing)

    def _completely_flatten(self, nplike, options):
        return self.project()._completely_flatten(nplike, options)

    def _recursively_apply(
        self, action, depth, depth_context, lateral_context, options
    ):
        if options["return_array"]:

            def continuation():
                return BitMaskedArray(
                    self._mask,
                    self._content._recursively_apply(
                        action,
                        depth,
                        copy.copy(depth_context),
                        lateral_context,
                        options,
                    ),
                    self._valid_when,
                    self._length,
                    self._lsb_order,
                    self._identifier,
                    self._parameters if options["keep_parameters"] else None,
                    self._nplike,
                )

        else:

            def continuation():
                self._content._recursively_apply(
                    action,
                    depth,
                    copy.copy(depth_context),
                    lateral_context,
                    options,
                )

        result = action(
            self,
            depth=depth,
            depth_context=depth_context,
            lateral_context=lateral_context,
            continuation=continuation,
            options=options,
        )

        if isinstance(result, Content):
            return result
        elif result is None:
            return continuation()
        else:
            raise AssertionError(result)

    def packed(self):
        if self._content.is_RecordType:
            next = self.toIndexedOptionArray64()

            content = next._content.packed()
            if content.length > self._length:
                content = content[: self._length]

            return ak._v2.contents.indexedoptionarray.IndexedOptionArray(
                next._index,
                content,
                next._identifier,
                next._parameters,
                self._nplike,
            )

        else:
            excess_length = int(math.ceil(self._length / 8.0))
            if self._mask.length == excess_length:
                mask = self._mask
            else:
                mask = self._mask[:excess_length]

            content = self._content.packed()
            if content.length > self._length:
                content = content[: self._length]

            return BitMaskedArray(
                mask,
                content,
                self._valid_when,
                self._length,
                self._lsb_order,
                self._identifier,
                self._parameters,
                self._nplike,
            )

    def _to_list(self, behavior):
        out = self._to_list_custom(behavior)
        if out is not None:
            return out

        mask = self.mask_as_bool(valid_when=True, nplike=self.nplike)[: self._length]
        content = self._content._to_list(behavior)
        out = [None] * self._length
        for i, isvalid in enumerate(mask):
            if isvalid:
                out[i] = content[i]
        return out

    def _to_nplike(self, nplike):
        content = self._content._to_nplike(nplike)
        mask = self._mask._to_nplike(nplike)
        return BitMaskedArray(
            mask,
            content,
            valid_when=self._valid_when,
            length=len(self),
            lsb_order=self._lsb_order,
            identifier=self._identifier,
            parameters=self._parameters,
            nplike=nplike,
        )

    def _to_json(
        self,
        nan_string,
        infinity_string,
        minus_infinity_string,
        complex_real_string,
        complex_imag_string,
    ):
        out = self._to_json_custom()
        if out is not None:
            return out

        mask = self.mask_as_bool(valid_when=True, nplike=self.nplike)[: self._length]
        content = self._content._to_json(
            nan_string,
            infinity_string,
            minus_infinity_string,
            complex_real_string,
            complex_imag_string,
        )
        out = [None] * self._length
        for i, isvalid in enumerate(mask):
            if isvalid:
                out[i] = content[i]
        return out



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE


import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


def strings_astype(array, to, highlevel=True, behavior=None):

    """
    Args:
        array: Array whose strings should be converted to a new numeric type.
        to (dtype or dtype specifier): Type to convert the strings into.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.layout.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.

    Converts all strings in the array to a new type, leaving the structure
    untouched.

    For example,

        >>> array = ak.Array(["1", "2", "    3    ", "00004", "-5"])
        >>> ak.strings_astype(array, np.int32)
        <Array [1, 2, 3, 4, -5] type='5 * int32'>

    and

        >>> array = ak.Array(["1.1", "2.2", "    3.3    ", "00004.4", "-5.5"])
        >>> ak.strings_astype(array, np.float64)
        <Array [1.1, 2.2, 3.3, 4.4, -5.5] type='5 * float64'>

    and finally,

        >>> array = ak.Array([["1.1", "2.2", "    3.3    "], [], ["00004.4", "-5.5"]])
        >>> ak.strings_astype(array, np.float64)
        <Array [[1.1, 2.2, 3.3], [], [4.4, -5.5]] type='3 * var * float64'>

    See also #ak.numbers_astype.
    """
    to_dtype = np.dtype(to)

    def action(layout, **kwargs):
        if layout.is_ListType and (
            layout.parameter("__array__") == "string"
            or layout.parameter("__array__") == "bytestring"
        ):
            layout = ak._v2.operations.structure.without_parameters(
                layout, highlevel=False
            )
            max_length = ak._v2.operations.reducers.max(
                ak._v2.operations.structure.num(layout)
            )
            regulararray = layout.rpad(max_length, 1)
            maskedarray = ak._v2.operations.convert.to_numpy(
                regulararray, allow_missing=True
            )
            npstrings = maskedarray.data
            if maskedarray.mask is not False:
                npstrings[maskedarray.mask] = 0
            npnumbers = (
                npstrings.reshape(-1).view("<S" + str(max_length)).astype(to_dtype)
            )
            return ak._v2.contents.NumpyArray(npnumbers)
        else:
            return None

    layout = ak._v2.operations.convert.to_layout(
        array, allow_record=False, allow_other=False
    )
    out = layout.recursively_apply(action)
    return ak._v2._util.wrap(out, behavior, highlevel)
